r"""
Isomorphisms between Weierstrass models of elliptic curves

AUTHORS:

- Robert Bradshaw (2007): initial version
- John Cremona (Jan 2008): isomorphisms, automorphisms and twists
  in all characteristics
- Lorenz Panny (2021): :class:`EllipticCurveHom` interface
"""
# ****************************************************************************
#   Copyright (C) 2007 Robert Bradshaw <robertwb@math.washington.edu>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from sage.structure.element import get_coercion_model

from .constructor import EllipticCurve
from sage.schemes.elliptic_curves.hom import EllipticCurveHom
from sage.structure.richcmp import (richcmp, richcmp_not_equal, op_EQ, op_NE)
from sage.structure.sequence import Sequence
from sage.rings.integer import Integer
from sage.rings.polynomial.polynomial_ring_constructor import PolynomialRing


class baseWI:
    r"""
    This class implements the basic arithmetic of isomorphisms between
    Weierstrass models of elliptic curves.

    These are specified by lists of the form `[u,r,s,t]` (with `u \neq 0`)
    which specifies a transformation `(x,y) \mapsto (x',y')` where

            `(x,y) = (u^2x'+r , u^3y' + su^2x' + t).`

    INPUT:

    - ``u``, ``r``, ``s``, ``t`` -- (default: `1`, `0`, `0`, `0`); standard
      parameters of an isomorphism between Weierstrass models

    EXAMPLES::

        sage: from sage.schemes.elliptic_curves.weierstrass_morphism import *
        sage: baseWI()
        (1, 0, 0, 0)
        sage: baseWI(2,3,4,5)
        (2, 3, 4, 5)
        sage: R.<u,r,s,t> = QQ[]
        sage: baseWI(u,r,s,t)
        (u, r, s, t)
    """
    def __init__(self, u=1, r=0, s=0, t=0):
        r"""
        Constructor: check for valid parameters (defaults to identity).

        INPUT:

        - ``u``, ``r``, ``s``, ``t`` -- (default: `1`, `0`, `0`, `0`); standard
          parameters of an isomorphism between Weierstrass models

        EXAMPLES::

            sage: from sage.schemes.elliptic_curves.weierstrass_morphism import *
            sage: baseWI()
            (1, 0, 0, 0)
            sage: baseWI(2,3,4,5)
            (2, 3, 4, 5)
            sage: R.<u,r,s,t> = QQ[]
            sage: baseWI(u,r,s,t)
            (u, r, s, t)
        """
        if not u:
            raise ValueError("u!=0 required for baseWI")
        self.u = u
        self.r = r
        self.s = s
        self.t = t

    def tuple(self):
        r"""
        Return the parameters `u,r,s,t` as a tuple.

        EXAMPLES::

            sage: from sage.schemes.elliptic_curves.weierstrass_morphism import *
            sage: w = baseWI(2,3,4,5)
            sage: w.tuple()
            (2, 3, 4, 5)
        """
        return (self.u, self.r, self.s, self.t)

    def __mul__(self, other):
        r"""
        Return the composition of this isomorphism and another.

        EXAMPLES::

            sage: from sage.schemes.elliptic_curves.weierstrass_morphism import *
            sage: baseWI(1,2,3,4)*baseWI(5,6,7,8)
            (5, 56, 22, 858)
            sage: baseWI()*baseWI(1,2,3,4)*baseWI()
            (1, 2, 3, 4)
        """
        u1, r1, s1, t1 = other.tuple()
        u2, r2, s2, t2 = self.tuple()
        return baseWI(u1 * u2,
                      (u1**2) * r2 + r1,
                      u1 * s2 + s1,
                      (u1**3) * t2 + s1 * (u1**2) * r2 + t1)

    def __invert__(self):
        r"""
        Return the inverse of this isomorphism.

        EXAMPLES::

            sage: from sage.schemes.elliptic_curves.weierstrass_morphism import *
            sage: w = baseWI(2,3,4,5)
            sage: ~w
            (1/2, -3/4, -2, 7/8)
            sage: w*~w
            (1, 0, 0, 0)
            sage: ~w*w
            (1, 0, 0, 0)
            sage: R.<u,r,s,t> = QQ[]
            sage: w = baseWI(u,r,s,t)
            sage: ~w
            (1/u, (-r)/u^2, (-s)/u, (r*s - t)/u^3)
            sage: ~w*w
            (1, 0, 0, 0)
        """
        u, r, s, t = self.tuple()
        return baseWI(1/u, -r/u**2, -s/u, (r*s-t)/u**3)

    def __repr__(self):
        r"""
        Return the string representation  of this isomorphism.

        EXAMPLES::

            sage: from sage.schemes.elliptic_curves.weierstrass_morphism import *
            sage: baseWI(2,3,4,5)
            (2, 3, 4, 5)
        """
        return repr(self.tuple())

    def is_identity(self):
        r"""
        Return ``True`` if this is the identity isomorphism.

        EXAMPLES::

            sage: from sage.schemes.elliptic_curves.weierstrass_morphism import *
            sage: w = baseWI(); w.is_identity()
            True
            sage: w = baseWI(2,3,4,5); w.is_identity()
            False
        """
        return self.tuple() == (1, 0, 0, 0)

    def __call__(self, EorP):
        r"""
        Base application of isomorphisms to curves and points.

        A baseWI `w` may be applied to a list `[a1,a2,a3,a4,a6]`
        representing the `a`-invariants of an elliptic curve `E`,
        returning the `a`-invariants of `w(E)`; or to `P=[x,y]` or
        `P=[x,y,z]` representing a point in `\mathbb{A}^2` or
        `\mathbb{P}^2`, returning the transformed point.

        INPUT:

        - ``EorP`` -- either an elliptic curve, or a point on an elliptic curve

        OUTPUT: the transformed curve or point

        EXAMPLES::

            sage: from sage.schemes.elliptic_curves.weierstrass_morphism import *
            sage: E = EllipticCurve([0,0,1,-7,6])
            sage: w = baseWI(2,3,4,5)
            sage: w(E.ainvs())
            [4, -7/4, 11/8, -3/2, -9/32]
            sage: P = E(-2,3)
            sage: w(P.xy())
            [-5/4, 9/4]
            sage: EllipticCurve(w(E.ainvs()))(w(P.xy()))
            (-5/4 : 9/4 : 1)
        """
        u, r, s, t = self.tuple()
        if len(EorP) == 5:
            a1, a2, a3, a4, a6 = EorP
            a6 += r*(a4 + r*(a2 + r)) - t*(a3 + r*a1 + t)
            a4 += -s*a3 + 2*r*a2 - (t + r*s)*a1 + 3*r*r - 2*s*t
            a3 += r*a1 + t + t
            a2 += -s*a1 + 3*r - s*s
            a1 += 2*s
            return [a1/u, a2/u**2, a3/u**3, a4/u**4, a6/u**6]
        if len(EorP) == 2:
            x, y = EorP
            x -= r
            y -= (s*x+t)
            return [x/u**2, y/u**3]
        if len(EorP) == 3:
            x, y, z = EorP
            x -= r*z
            y -= (s*x+t*z)
            return [x/u**2, y/u**3, z]
        raise ValueError("baseWI(a) only for a=(x,y), (x:y:z) or (a1,a2,a3,a4,a6)")


def _isomorphisms(E, F):
    r"""
    Enumerate all isomorphisms between two elliptic curves,
    as a generator object.

    INPUT:

    - ``E``, ``F`` -- two elliptic curves

    OUTPUT:

    A generator object producing 4-tuples `(u,r,s,t)` representing an isomorphism.

    EXAMPLES::

        sage: from sage.schemes.elliptic_curves.weierstrass_morphism import _isomorphisms
        sage: list(_isomorphisms(EllipticCurve_from_j(0), EllipticCurve('27a3')))
        [(1, 0, 0, 0), (-1, 0, 0, -1)]
        sage: list(_isomorphisms(EllipticCurve_from_j(0), EllipticCurve('27a1')))
        []

    TESTS:

    Check that :issue:`32632` is fixed::

        sage: z8 = GF(2^8).gen()
        sage: E1 = EllipticCurve([z8, z8, z8, z8, z8])
        sage: list(_isomorphisms(E1, E1))
        [(1, 0, 0, 0), (1, 0, z8, z8)]
        sage: E2 = EllipticCurve([z8^2, 0, 0, 0, z8^7 + z8^4])
        sage: list(_isomorphisms(E1, E2))
        [(z8^7 + z8^3 + z8^2 + z8, 1, 1, z8^7 + z8^3 + z8^2 + z8 + 1),
         (z8^7 + z8^3 + z8^2 + z8, 1, z8 + 1, z8^7 + z8^3 + z8^2 + z8 + 1)]

    Random testing::

        sage: p = random_prime(100)
        sage: F = GF(p).algebraic_closure()
        sage: j = F.random_element()
        sage: while j in (0, 1728):     # skip the hard case
        ....:     j = F.random_element()
        sage: j = F(choice((0, 1728)))  # long time -- do the hard case
        sage: E = EllipticCurve_from_j(j)
        sage: u,r,s,t = (F^4).random_element()
        sage: u = u or 1
        sage: E = E.change_weierstrass_model(u,r,s,t)
        sage: Aut = E.automorphisms()
        sage: len(set(Aut)) == len(Aut)
        True
        sage: all(-a in Aut for a in Aut)
        True
        sage: len(Aut) in (2, 4, 6, 12, 24)
        True
        sage: j = E.j_invariant()
        sage: {
        ....:      2: j not in (0, 1728),
        ....:      4: p >= 5 and j == 1728,
        ....:      6: p >= 5 and j == 0,
        ....:     12: p == 3 and j == 0,  # note 1728 == 0
        ....:     24: p == 2 and j == 0,  # note 1728 == 0
        ....: }[len(Aut)]
        True
        sage: u,r,s,t = (F^4).random_element()
        sage: u = u or 1
        sage: F = E.change_weierstrass_model(u,r,s,t)
        sage: Iso = E.isomorphisms(F)
        sage: len(set(Iso)) == len(Iso)
        True
        sage: all(-f in Iso for f in Iso)
        True
        sage: len(Iso) == len(Aut)
        True
        sage: all({iso2*iso1 for iso1 in Iso} == set(Aut) for iso2 in F.isomorphisms(E))
        True
    """
    from .ell_generic import EllipticCurve_generic
    if not isinstance(E, EllipticCurve_generic) or not isinstance(F, EllipticCurve_generic):
        raise ValueError("arguments are not elliptic curves")

    j = E.j_invariant()
    if j != F.j_invariant():
        return

    K = E.base_ring()

    from sage.rings.polynomial.polynomial_ring import polygen
    x = polygen(K, 'x')

    a1E, a2E, a3E, a4E, a6E = E.ainvs()
    a1F, a2F, a3F, a4F, a6F = F.ainvs()

    char = K.characteristic()

    if char == 2:
        if j == 0:
            ulist = (x**3 - a3E/a3F).roots(multiplicities=False)
            for u in ulist:
                slist = (x**4 + a3E*x + (a2F**2 + a4F)*u**4 + a2E**2 + a4E).roots(multiplicities=False)
                for s in slist:
                    r = s**2 + a2E + a2F*u**2
                    tlist = (x**2 + a3E*x + r**3 + a2E*r**2 + a4E*r + a6E + a6F*u**6).roots(multiplicities=False)
                    for t in tlist:
                        yield (u, r, s, t)
        else:
            u = a1E/a1F
            r = (a3E + a3F*u**3)/a1E
            slist = (x**2 + a1E*x + r + a2E + a2F*u**2).roots(multiplicities=False)
            for s in slist:
                t = (a4E + a4F*u**4 + s*a3E + r*s*a1E + r**2) / a1E
                yield (u, r, s, t)
        return

    b2E, b4E, b6E, b8E = E.b_invariants()
    b2F, b4F, b6F, b8F = F.b_invariants()

    if char == 3:
        if j == 0:
            ulist = (x**4 - b4E/b4F).roots(multiplicities=False)
            for u in ulist:
                s = a1E - a1F*u
                t = a3E - a3F*u**3
                rlist = (x**3 - b4E*x + b6E - b6F*u**6).roots(multiplicities=False)
                for r in rlist:
                    yield (u, r, s, t + r*a1E)
        else:
            ulist = (x**2 - b2E/b2F).roots(multiplicities=False)
            for u in ulist:
                r = (b4F * u**4 - b4E) / b2E
                s = a1E - a1F * u
                t = a3E - a3F * u**3 + a1E * r
                yield (u, r, s, t)
        return

    # now char != 2,3:

    c4E, c6E = E.c_invariants()
    c4F, c6F = F.c_invariants()

    if j == 0:
        m, um = 6, c6E/c6F
    elif j == 1728:
        m, um = 4, c4E/c4F
    else:
        m, um = 2, (c6E*c4F)/(c6F*c4E)
    ulist = (x**m - um).roots(multiplicities=False)
    for u in ulist:
        s = (a1F*u - a1E)/2
        r = (a2F*u**2 + a1E*s + s**2 - a2E)/3
        t = (a3F*u**3 - a1E*r - a3E)/2
        yield (u, r, s, t)


class WeierstrassIsomorphism(EllipticCurveHom, baseWI):
    r"""
    Class representing a Weierstrass isomorphism between two elliptic curves.

    INPUT:

    - ``E`` -- an ``EllipticCurve``, or ``None`` (see below)

    - ``urst`` -- a 4-tuple `(u,r,s,t)`, a :class:`baseWI` object,
      or ``None`` (see below)

    - ``F`` -- an ``EllipticCurve``, or ``None`` (see below)

    Given two Elliptic Curves ``E`` and ``F`` (represented by Weierstrass
    models as usual), and a transformation ``urst`` from ``E`` to ``F``,
    construct an isomorphism from ``E`` to ``F``.
    An exception is raised if ``urst(E) != F``.  At most one of ``E``,
    ``F``, ``urst`` can be ``None``.  In this case, the missing input is
    constructed from the others in such a way that ``urst(E) == F`` holds,
    and an exception is raised if this is impossible (typically because
    ``E`` and ``F`` are not isomorphic).

    Users will not usually need to use this class directly, but instead use
    methods such as
    :meth:`~sage.schemes.elliptic_curves.ell_generic.EllipticCurve_generic.isomorphism_to`
    or
    :meth:`~sage.schemes.elliptic_curves.ell_generic.EllipticCurve_generic.isomorphisms`.

    Explicitly, the isomorphism defined by `(u,r,s,t)` maps a point `(x,y)`
    to the point

    .. MATH::

        ((x-r) / u^2, \; (y - s(x-r) - t) / u^3) .

    If the domain `E` has Weierstrass coefficients `[a_1,a_2,a_3,a_4,a_6]`,
    the codomain `F` is given by

    .. MATH::

        a_1' &= (a_1 + 2s) / u \\
        a_2' &= (a_2 - a_1s + 3r - s^2) / u^2 \\
        a_3' &= (a_3 + a_1r + 2t) / u^3 \\
        a_4' &= (a_4 + 2a_2r - a_1(rs+t) - a_3s + 3r^2 - 2st) / u^4 \\
        a_6' &= (a_6 - a_1rt + a_2r^2 - a_3t + a_4r + r^3 - t^2) / u^6 .

    EXAMPLES::

        sage: from sage.schemes.elliptic_curves.weierstrass_morphism import *
        sage: WeierstrassIsomorphism(EllipticCurve([0,1,2,3,4]), (-1,2,3,4))
        Elliptic-curve morphism:
          From: Elliptic Curve defined by y^2 + 2*y = x^3 + x^2 + 3*x + 4 over Rational Field
          To:   Elliptic Curve defined by y^2 - 6*x*y - 10*y = x^3 - 2*x^2 - 11*x - 2 over Rational Field
          Via:  (u,r,s,t) = (-1, 2, 3, 4)
        sage: E = EllipticCurve([0,1,2,3,4])
        sage: F = EllipticCurve(E.cremona_label())
        sage: WeierstrassIsomorphism(E, None, F)
        Elliptic-curve morphism:
          From: Elliptic Curve defined by y^2 + 2*y = x^3 + x^2 + 3*x + 4 over Rational Field
          To:   Elliptic Curve defined by y^2  = x^3 + x^2 + 3*x + 5 over Rational Field
          Via:  (u,r,s,t) = (1, 0, 0, -1)
        sage: w = WeierstrassIsomorphism(None, (1,0,0,-1), F)
        sage: w._domain == E
        True
    """
    def __init__(self, E=None, urst=None, F=None):
        r"""
        Constructor for the ``WeierstrassIsomorphism`` class.

        TESTS:

        Check for :issue:`33215`::

            sage: from sage.schemes.elliptic_curves.weierstrass_morphism import WeierstrassIsomorphism
            sage: E = EllipticCurve(GF(71^2), [5,5])                                    # needs sage.rings.finite_rings
            sage: iso = WeierstrassIsomorphism(E, (1,2,3,4))                            # needs sage.rings.finite_rings
            sage: ~iso  # indirect doctest                                              # needs sage.rings.finite_rings
            Elliptic-curve morphism:
              From: Elliptic Curve defined by y^2 + 6*x*y + 8*y = x^3 + 68*x^2 + 64*x + 7 over Finite Field in z2 of size 71^2
              To:   Elliptic Curve defined by y^2 = x^3 + 5*x + 5 over Finite Field in z2 of size 71^2
              Via:  (u,r,s,t) = (1, 69, 68, 2)

        Test for :issue:`33312`::

            sage: type(iso.degree())                                                    # needs sage.rings.finite_rings
            <class 'sage.rings.integer.Integer'>
        """
        from .ell_generic import EllipticCurve_generic

        if E is not None:
            if not isinstance(E, EllipticCurve_generic):
                raise ValueError("first argument must be an elliptic curve or None")
        if F is not None:
            if not isinstance(F, EllipticCurve_generic):
                raise ValueError("third argument must be an elliptic curve or None")
        if urst is not None:
            if len(urst) != 4:
                raise ValueError("second argument must be [u,r,s,t] or None")
        if len([par for par in [E, urst, F] if par is not None]) < 2:
            raise ValueError("at most 1 argument can be None")

        inps = []
        if E is not None:
            inps.append(E.base_ring())
        if F is not None:
            inps.append(F.base_ring())
        if urst is not None:
            inps += list(urst)
        base_ring = get_coercion_model().common_parent(*inps)

        if urst is not None:
            urst = Sequence(urst, base_ring)

        if F is None:  # easy case
            baseWI.__init__(self, *urst)
            F = EllipticCurve(baseWI.__call__(self, list(E.a_invariants())))

        elif E is None:  # easy case in reverse
            baseWI.__init__(self, *urst)
            inv_urst = baseWI.__invert__(self)
            E = EllipticCurve(baseWI.__call__(inv_urst, list(F.a_invariants())))

        elif urst is None:  # try to construct the morphism
            try:
                urst = next(_isomorphisms(E, F))
            except StopIteration:
                raise ValueError("elliptic curves not isomorphic")
            baseWI.__init__(self, *urst)

        else:  # none of the parameters is None:
            baseWI.__init__(self, *urst)
            if F != EllipticCurve(baseWI.__call__(self, list(E.a_invariants()))):
                raise ValueError("second argument is not an isomorphism from first argument to third argument")

        self._mpoly_ring = PolynomialRing(base_ring, ['x','y'])
        self._poly_ring = PolynomialRing(base_ring, ['x'])
        self._xyfield = self._mpoly_ring.fraction_field()
        self._xfield = self._poly_ring.fraction_field()

        self._domain = E
        self._codomain = F
        self._degree = Integer(1)
        EllipticCurveHom.__init__(self, self._domain, self._codomain)

    @staticmethod
    def _comparison_impl(left, right, op):
        r"""
        Compare an isomorphism to another elliptic-curve morphism.

        Called by :meth:`EllipticCurveHom._richcmp_`.

        EXAMPLES::

            sage: from sage.schemes.elliptic_curves.weierstrass_morphism import *
            sage: E = EllipticCurve('389a1')
            sage: F = E.change_weierstrass_model(1,2,3,4)
            sage: w1 = E.isomorphism_to(F)
            sage: w1 == w1
            True
            sage: w2 = F.automorphisms()[1] * w1
            sage: w1 == w2
            False

            sage: E = EllipticCurve_from_j(GF(7)(0))
            sage: F = E.change_weierstrass_model(2,3,4,5)
            sage: a = E.isomorphisms(F)
            sage: b = [w*a[0] for w in F.automorphisms()]
            sage: b.sort()
            sage: a == b
            True
            sage: c = [a[0]*w for w in E.automorphisms()]
            sage: c.sort()
            sage: a == c
            True
        """
        if not isinstance(left, WeierstrassIsomorphism) or not isinstance(right, WeierstrassIsomorphism):
            return NotImplemented

        lx = left._domain
        rx = right._domain
        if lx != rx:
            return richcmp_not_equal(lx, rx, op)

        lx = left._codomain
        rx = right._codomain
        if lx != rx:
            return richcmp_not_equal(lx, rx, op)

        if op in (op_EQ, op_NE):
            return richcmp(left.tuple(), right.tuple(), op)

        # This makes sure that the identity and negation morphisms
        # come first in a sorted list of WeierstrassIsomorphisms.
        # More generally, we're making sure that a morphism and its
        # negative appear next to each other, and that those pairs
        # of isomorphisms satisfying u=+-1 come first.
        def _sorting_key(iso):
            v, w = iso.tuple(), (-iso).tuple()
            i = 0 if (1,0,0,0) in (v,w) else 1
            j = 0 if v[0] == 1 else 1 if w[0] == 1 else 2
            return (i,) + min(v,w) + (j,) + v

        return richcmp(_sorting_key(left), _sorting_key(right), op)

    def _eval(self, P):
        r"""
        Less strict evaluation method for internal use.

        In particular, this can be used to evaluate ``self`` at a
        point defined over an extension field.

        INPUT: a sequence of 3 coordinates defining a point on ``self``

        OUTPUT: the result of evaluating ``self`` at the given point

        EXAMPLES::

            sage: # needs sage.rings.number_field
            sage: from sage.schemes.elliptic_curves.weierstrass_morphism import WeierstrassIsomorphism
            sage: E = EllipticCurve([i, 0]); E
            Elliptic Curve defined by y^2 = x^3 + I*x
             over Number Field in I with defining polynomial x^2 + 1 with I = 1*I
            sage: iso = WeierstrassIsomorphism(E, (i,1,2,3))
            sage: P = E.change_ring(QQbar).lift_x(QQbar.random_element())
            sage: Q = iso._eval(P)
            sage: Q.curve()
            Elliptic Curve defined by y^2 + (-4*I)*x*y + 6*I*y = x^3 + x^2 + (I-9)*x + (-I+8)
             over Algebraic Field
            sage: y = next(filter(bool, iter(QQbar.random_element, None)))  # sample until nonzero
            sage: iso._eval((0, y, 0)) == 0
            True
        """
        if self._domain.defining_polynomial()(*P):
            raise ValueError(f'{P} not on {self._domain}')
        k = Sequence(P).universe()

        Q = baseWI.__call__(self, P)
        return self._codomain.base_extend(k).point(Q)

    def _call_(self, P):
        r"""
        Call function for WeierstrassIsomorphism class.

        INPUT:

        - ``P`` -- Point; a point on the domain curve

        OUTPUT:

        (Point) the transformed point on the codomain curve.

        EXAMPLES::

            sage: from sage.schemes.elliptic_curves.weierstrass_morphism import *
            sage: E = EllipticCurve('37a1')
            sage: w = WeierstrassIsomorphism(E,(2,3,4,5))
            sage: P = E(0,-1)
            sage: w(P)
            (-3/4 : 3/4 : 1)
            sage: w(P).curve() == E.change_weierstrass_model((2,3,4,5))
            True

        TESTS:

        Check that copying the order over works::

            sage: # needs sage.rings.finite_rings
            sage: E = EllipticCurve(GF(431^2), [1,0])
            sage: i = next(a for a in E.automorphisms() if a^2 == -a^24)
            sage: P,_ = E.gens()
            sage: P._order
            432
            sage: i(P)._order
            432
            sage: E(i(P))._order
            432

        Check that the isomorphism cannot be evaluated on points outside
        its domain (see :issue:`35799`)::

            sage: # needs sage.rings.finite_rings
            sage: E = EllipticCurve(GF(101), [1,1])
            sage: f = E.automorphisms()[0]
            sage: EE = EllipticCurve(GF(101), [5,5])
            sage: P = EE.lift_x(2)
            sage: P in f.domain()
            False
            sage: f(P)
            Traceback (most recent call last):
            ...
            TypeError: (2 : 15 : 1) fails to convert into the map's
            domain Elliptic Curve defined by y^2 = x^3 + x + 1 over
            Finite Field of size 101, but a `pushforward` method is
            not properly implemented
        """
        if P[2] == 0:
            return self._codomain(0)
        res = baseWI.__call__(self, tuple(P._coords))
        Q = self._codomain.point(res, check=False)
        if hasattr(P, '_order'):
            Q._order = P._order
        return Q

    def __invert__(self):
        r"""
        Return the inverse of this WeierstrassIsomorphism.

        EXAMPLES::

            sage: E = EllipticCurve('5077')
            sage: F = E.change_weierstrass_model([2,3,4,5]); F
            Elliptic Curve defined by y^2 + 4*x*y + 11/8*y = x^3 - 7/4*x^2 - 3/2*x - 9/32 over Rational Field
            sage: w = E.isomorphism_to(F)
            sage: P = E(-2,3,1)
            sage: w(P)
            (-5/4 : 9/4 : 1)
            sage: ~w
            Elliptic-curve morphism:
              From: Elliptic Curve defined by y^2 + 4*x*y + 11/8*y = x^3 - 7/4*x^2 - 3/2*x - 9/32 over Rational Field
              To:   Elliptic Curve defined by y^2 + y = x^3 - 7*x + 6 over Rational Field
              Via:  (u,r,s,t) = (1/2, -3/4, -2, 7/8)
            sage: Q = w(P); Q
            (-5/4 : 9/4 : 1)
            sage: (~w)(Q)
            (-2 : 3 : 1)
        """
        winv = baseWI.__invert__(self).tuple()
        return WeierstrassIsomorphism(self._codomain, winv, self._domain)

    @staticmethod
    def _composition_impl(left, right):
        r"""
        Return the composition of a ``WeierstrassIsomorphism``
        with another elliptic-curve morphism.

        Called by :meth:`EllipticCurveHom._composition_`.

        EXAMPLES::

            sage: E1 = EllipticCurve('5077')
            sage: E2 = E1.change_weierstrass_model([2,3,4,5])
            sage: w1 = E1.isomorphism_to(E2)
            sage: E3 = E2.change_weierstrass_model([6,7,8,9])
            sage: w2 = E2.isomorphism_to(E3)
            sage: P = E1(-2,3,1)
            sage: (w2*w1)(P) == w2(w1(P))
            True

        TESTS:

        We should return ``NotImplemented`` when passed a combination of
        elliptic-curve morphism types that we don't handle here::

            sage: E1 = EllipticCurve([1,0])
            sage: phi = E1.isogeny(E1(0,0))
            sage: E2 = phi.codomain()
            sage: psi = E2.isogeny(E2(0,0))
            sage: w1._composition_impl(psi, phi)
            NotImplemented
        """
        if isinstance(left, WeierstrassIsomorphism) and isinstance(right, WeierstrassIsomorphism):
            if left._domain != right._codomain:
                raise ValueError("Domain of first argument must equal codomain of second")
            w = baseWI.__mul__(left, right)
            return WeierstrassIsomorphism(right._domain, w.tuple(), left._codomain)

        return NotImplemented

    def __repr__(self):
        r"""
        Return the string representation of this WeierstrassIsomorphism.

        OUTPUT:

        (string) The underlying morphism, together with an extra line
        showing the `(u,r,s,t)` parameters.

        EXAMPLES::

            sage: E1 = EllipticCurve('5077')
            sage: E2 = E1.change_weierstrass_model([2,3,4,5])
            sage: E1.isomorphism_to(E2)
            Elliptic-curve morphism:
              From: Elliptic Curve defined by y^2 + y = x^3 - 7*x + 6 over Rational Field
              To:   Elliptic Curve defined by y^2 + 4*x*y + 11/8*y = x^3 - 7/4*x^2 - 3/2*x - 9/32 over Rational Field
              Via:  (u,r,s,t) = (2, 3, 4, 5)
        """
        return EllipticCurveHom.__repr__(self) + "\n  Via:  (u,r,s,t) = " + baseWI.__repr__(self)

    # EllipticCurveHom methods

    def rational_maps(self):
        """
        Return the pair of rational maps defining this isomorphism.

        EXAMPLES::

            sage: E1 = EllipticCurve([11,22,33,44,55])
            sage: E2 = EllipticCurve_from_j(E1.j_invariant())
            sage: iso = E1.isomorphism_to(E2); iso
            Elliptic-curve morphism:
              From: Elliptic Curve defined by y^2 + 11*x*y + 33*y = x^3 + 22*x^2 + 44*x + 55
                    over Rational Field
              To:   Elliptic Curve defined by y^2 + x*y = x^3 + x^2 - 684*x + 6681
                    over Rational Field
              Via:  (u,r,s,t) = (1, -17, -5, 77)
            sage: iso.rational_maps()
            (x + 17, 5*x + y + 8)
            sage: f = E2.defining_polynomial()(*iso.rational_maps(), 1)
            sage: I = E1.defining_ideal()
            sage: x,y,z = I.ring().gens()
            sage: f in I + Ideal(z-1)
            True

        ::

            sage: # needs sage.rings.finite_rings
            sage: E = EllipticCurve(GF(65537), [1,1,1,1,1])
            sage: w = E.isomorphism_to(E.short_weierstrass_model())
            sage: f,g = w.rational_maps()
            sage: P = E.random_point()
            sage: w(P).xy() == (f(P.xy()), g(P.xy()))
            True

        TESTS:

        Check for :issue:`34811`::

            sage: iso.rational_maps()[0].parent()
            Fraction Field of Multivariate Polynomial Ring in x, y over Rational Field
            sage: iso.rational_maps()[1].parent()
            Fraction Field of Multivariate Polynomial Ring in x, y over Rational Field
        """
        return tuple(baseWI.__call__(self, self._xyfield.gens()))

    def x_rational_map(self):
        """
        Return the `x`-coordinate rational map of this isomorphism.

        EXAMPLES::

            sage: E1 = EllipticCurve([11,22,33,44,55])
            sage: E2 = EllipticCurve_from_j(E1.j_invariant())
            sage: iso = E1.isomorphism_to(E2); iso
            Elliptic-curve morphism:
              From: Elliptic Curve defined by y^2 + 11*x*y + 33*y = x^3 + 22*x^2 + 44*x + 55
                    over Rational Field
              To:   Elliptic Curve defined by y^2 + x*y = x^3 + x^2 - 684*x + 6681
                    over Rational Field
              Via:  (u,r,s,t) = (1, -17, -5, 77)
            sage: iso.x_rational_map()
            x + 17
            sage: iso.x_rational_map() == iso.rational_maps()[0]
            True

        TESTS:

        Check for :issue:`34811`::

            sage: iso.x_rational_map().parent()
            Fraction Field of Univariate Polynomial Ring in x over Rational Field
        """
        x, = self._xfield.gens()
        return (x - self.r) / self.u**2

    def kernel_polynomial(self):
        """
        Return the kernel polynomial of this isomorphism.

        Isomorphisms have trivial kernel by definition, hence this
        method always returns `1`.

        EXAMPLES::

            sage: E1 = EllipticCurve([11,22,33,44,55])
            sage: E2 = EllipticCurve_from_j(E1.j_invariant())
            sage: iso = E1.isomorphism_to(E2)
            sage: iso.kernel_polynomial()
            1
            sage: psi = E1.isogeny(iso.kernel_polynomial(), codomain=E2); psi
            Isogeny of degree 1
             from Elliptic Curve defined by y^2 + 11*x*y + 33*y = x^3 + 22*x^2 + 44*x + 55
                  over Rational Field
               to Elliptic Curve defined by y^2 + x*y = x^3 + x^2 - 684*x + 6681
                  over Rational Field
            sage: psi in {iso, -iso}
            True

        TESTS::

            sage: iso.kernel_polynomial().parent()
            Univariate Polynomial Ring in x over Rational Field
        """
        return self._poly_ring(1)

    def dual(self):
        """
        Return the dual isogeny of this isomorphism.

        For isomorphisms, the dual is just the inverse.

        EXAMPLES::

            sage: from sage.schemes.elliptic_curves.weierstrass_morphism import WeierstrassIsomorphism
            sage: E = EllipticCurve(QuadraticField(-3), [0,1])                          # needs sage.rings.number_field
            sage: w = WeierstrassIsomorphism(E, (CyclotomicField(3).gen(),0,0,0))       # needs sage.rings.number_field
            sage: (w.dual() * w).rational_maps()                                        # needs sage.rings.number_field
            (x, y)

        ::

            sage: E1 = EllipticCurve([11,22,33,44,55])
            sage: E2 = E1.short_weierstrass_model()
            sage: iso = E1.isomorphism_to(E2)
            sage: iso.dual() == ~iso
            True
        """
        return ~self

    def __neg__(self):
        """
        Return the negative of this isomorphism, i.e., its composition
        with the negation map `[-1]`.

        EXAMPLES::

            sage: from sage.schemes.elliptic_curves.weierstrass_morphism import WeierstrassIsomorphism
            sage: E = EllipticCurve([11,22,33,44,55])
            sage: w = WeierstrassIsomorphism(E, (66,77,88,99))
            sage: -w
            Elliptic-curve morphism:
              From: Elliptic Curve defined by y^2 + 11*x*y + 33*y = x^3 + 22*x^2 + 44*x + 55 over Rational Field
              To:   Elliptic Curve defined by y^2 + 17/6*x*y + 49/13068*y = x^3 - 769/396*x^2 - 3397/862488*x + 44863/7513995456
                    over Rational Field
              Via:  (u,r,s,t) = (-66, 77, -99, -979)
            sage: -(-w) == w
            True

        ::

            sage: # needs sage.rings.number_field
            sage: from sage.schemes.elliptic_curves.weierstrass_morphism import WeierstrassIsomorphism
            sage: K.<a> = QuadraticField(-3)
            sage: E = EllipticCurve(K, [0,1])
            sage: w = WeierstrassIsomorphism(E, (CyclotomicField(3).gen(),0,0,0))
            sage: w.tuple()
            (1/2*a - 1/2, 0, 0, 0)
            sage: (-w).tuple()
            (-1/2*a + 1/2, 0, 0, 0)
            sage: (-w)^3 == -(w^3)
            True

        ::

            sage: from sage.schemes.elliptic_curves.weierstrass_morphism import WeierstrassIsomorphism, identity_morphism
            sage: E = EllipticCurve(QuadraticField(-1), [1,0])                          # needs sage.rings.number_field
            sage: t = WeierstrassIsomorphism(E, (i,0,0,0))                              # needs sage.rings.number_field
            sage: -t^2 == identity_morphism(E)                                          # needs sage.rings.number_field
            True
        """
        a1,_,a3,_,_ = self._domain.a_invariants()
        w = baseWI(-1, 0, -a1, -a3)
        urst = baseWI.__mul__(self, w).tuple()
        return WeierstrassIsomorphism(self._domain, urst, self._codomain)

    def scaling_factor(self):
        r"""
        Return the Weierstrass scaling factor associated to this
        Weierstrass isomorphism.

        The scaling factor is the constant `u` (in the base field)
        such that `\varphi^* \omega_2 = u \omega_1`, where
        `\varphi: E_1\to E_2` is this isomorphism and `\omega_i` are
        the standard Weierstrass differentials on `E_i` defined by
        `\mathrm dx/(2y+a_1x+a_3)`.

        EXAMPLES::

            sage: E = EllipticCurve(QQbar, [0,1])                                       # needs sage.rings.number_field
            sage: all(f.scaling_factor() == f.formal()[1] for f in E.automorphisms())   # needs sage.rings.number_field
            True

        ALGORITHM: The scaling factor equals the `u` component of
        the tuple `(u,r,s,t)` defining the isomorphism.
        """
        return self.u

    def inseparable_degree(self):
        r"""
        Return the inseparable degree of this Weierstrass isomorphism.

        For isomorphisms, this method always returns one.

        TESTS::

            sage: from sage.schemes.elliptic_curves.weierstrass_morphism import WeierstrassIsomorphism
            sage: WeierstrassIsomorphism.inseparable_degree(None)
            1
        """
        return Integer(1)

    def is_identity(self):
        r"""
        Check if this Weierstrass isomorphism is the identity.

        EXAMPLES::

            sage: from sage.schemes.elliptic_curves.weierstrass_morphism import WeierstrassIsomorphism
            sage: p = 97
            sage: Fp = GF(p)
            sage: E = EllipticCurve(Fp, [1, 28])
            sage: ws = WeierstrassIsomorphism(E, None, E)
            sage: ws.is_identity()
            False

        ::

            sage: from sage.schemes.elliptic_curves.weierstrass_morphism import WeierstrassIsomorphism
            sage: p = 97
            sage: Fp = GF(p)
            sage: E = EllipticCurve(Fp, [1, 28])
            sage: ws = WeierstrassIsomorphism(E, (1, 0, 0, 0), None)
            sage: ws.is_identity()
            True
        """
        return self.tuple() == (1, 0, 0, 0)

    def order(self):
        r"""
        Compute the order of this Weierstrass isomorphism if it is an automorphism.

        A :exc:`ValueError` is raised if the domain is not equal to the codomain.

        A :exc:`NotImplementedError` is raised if the order of the automorphism is not 1, 2, 3, 4 or 6.

        EXAMPLES::

            sage: from sage.schemes.elliptic_curves.weierstrass_morphism import *
            sage: p = 97
            sage: Fp = GF(p)
            sage: E = EllipticCurve(Fp, [1, 28])
            sage: ws = WeierstrassIsomorphism(E, None, E)
            sage: ws.order()
            2

        TESTS::

            sage: from sage.schemes.elliptic_curves.weierstrass_morphism import *
            sage: p = 97
            sage: Fp = GF(p)
            sage: E = EllipticCurve(Fp, [1, 28])
            sage: ws = WeierstrassIsomorphism(E, None, E)
            sage: ws.order()
            2
            sage: E1 = EllipticCurve(Fp, [1, 69])
            sage: ws = E.isomorphism_to(E1)
            sage: ws.order()
            Traceback (most recent call last):
            ...
            ValueError: the domain is different from the codomain

        ::

            sage: E = EllipticCurve_from_j(Fp(0))
            sage: ws = WeierstrassIsomorphism(E, (Fp(36), 0, 0, 0), None)
            sage: ws.order()
            6
            sage: ws2 = ws*ws
            sage: ws2.order()
            3
            sage: F2_bar = GF(2).algebraic_closure()
            sage: E = EllipticCurve_from_j(F2_bar(0))
            sage: ws = WeierstrassIsomorphism(E, None, E)
            sage: ws.order()
            3
        """
        # Check if it is an actual endomorphism
        if self._domain != self._codomain:
            raise ValueError("the domain is different from the codomain")

        if self.is_identity():
            return Integer(1)

        ws2 = WeierstrassIsomorphism._composition_impl(self, self)
        if ws2.is_identity():
            return Integer(2)

        ws3 = WeierstrassIsomorphism._composition_impl(self, ws2)
        if ws3.is_identity():
            return Integer(3)

        ws4 = WeierstrassIsomorphism._composition_impl(ws2, ws2)
        if ws4.is_identity():
            return Integer(4)

        ws6 = WeierstrassIsomorphism._composition_impl(ws2, ws4)
        if ws6.is_identity():
            return Integer(6)

        raise NotImplementedError("the order of the endomorphism is not 1, 2, 3, 4 or 6")


def identity_morphism(E):
    r"""
    Given an elliptic curve `E`, return the identity morphism
    on `E` as a :class:`WeierstrassIsomorphism`.

    EXAMPLES::

        sage: from sage.schemes.elliptic_curves.weierstrass_morphism import identity_morphism
        sage: E = EllipticCurve([5,6,7,8,9])
        sage: id_ = identity_morphism(E)
        sage: id_.rational_maps()
        (x, y)
    """
    R = E.base_ring()
    zero = R.zero()
    return WeierstrassIsomorphism(E, (R.one(), zero, zero, zero))


def negation_morphism(E):
    r"""
    Given an elliptic curve `E`, return the negation endomorphism
    `[-1]` of `E` as a :class:`WeierstrassIsomorphism`.

    EXAMPLES::

        sage: from sage.schemes.elliptic_curves.weierstrass_morphism import negation_morphism
        sage: E = EllipticCurve([5,6,7,8,9])
        sage: neg = negation_morphism(E)
        sage: neg.rational_maps()
        (x, -5*x - y - 7)
    """
    R = E.base_ring()
    return WeierstrassIsomorphism(E, (-R.one(), R.zero(), -E.a1(), -E.a3()))



from pathlib import Path

import sage

VERSION = "1.2.3"

# The following must not be used during build to determine source or installation
# location of sagelib.  See comments in SAGE_ROOT/src/Makefile.in
# These variables come first so that other substituted variable values can refer
# to it.
SAGE_LOCAL = "/usr"

# The semicolon-separated list of GAP root paths. This is the list of
# locations that are searched for GAP packages. This is passed directly
# to GAP via the -l flag.
GAP_ROOT_PATHS = "/usr/lib/gap/;/usr/share/gap/".replace("${prefix}", SAGE_LOCAL)

# The path to the standalone maxima executable.
MAXIMA = "/usr/bin/maxima".replace("${prefix}", SAGE_LOCAL)

# Set this to the empty string if your ECL can load maxima without
# further prodding.
MAXIMA_FAS = "".replace("${prefix}", SAGE_LOCAL)
MAXIMA_PREFIX = "".replace("${prefix}", SAGE_LOCAL)

# Delete this line if your ECL can load Kenzo without further prodding.
KENZO_FAS = "'${prefix}'/lib/ecl/kenzo.fas".replace("${prefix}", SAGE_LOCAL)

NTL_INCDIR = ""
NTL_LIBDIR = ""

# Path to the ecl-config script
ECL_CONFIG = "/usr/bin/ecl-config".replace("${prefix}", SAGE_LOCAL)

# Path to the nauty binaries; of the form "/path/to/" or "/path/to/nauty-" 
SAGE_NAUTY_BINS_PREFIX = "/usr/bin/"

SAGE_ECMBIN = "/usr/bin/ecm"

# Names or paths of the 4ti2 executables
FOURTITWO_HILBERT = ""
FOURTITWO_MARKOV = ""
FOURTITWO_GRAVER = ""
FOURTITWO_ZSOLVE = ""
FOURTITWO_QSOLVE = ""
FOURTITWO_RAYS = ""
FOURTITWO_PPI = ""
FOURTITWO_CIRCUITS = ""
FOURTITWO_GROEBNER = ""

# for sage_setup.setenv
SAGE_ARCHFLAGS = "unset"
SAGE_PKG_CONFIG_PATH = "".replace("$SAGE_LOCAL", SAGE_LOCAL)

# Used in sage.repl.ipython_kernel.install
MATHJAX_DIR = "".replace("${prefix}", SAGE_LOCAL)
THREEJS_DIR = SAGE_LOCAL + "/share/threejs-sage"

# OpenMP flags, if available.
OPENMP_CFLAGS = "-fopenmp"
OPENMP_CXXFLAGS = "-fopenmp"


def is_editable_install() -> bool:
    """
    Check whether this is an editable install of Sage.

    EXAMPLES::

        sage: from sage.config import is_editable_install
        sage: is_editable_install()
        False
    """
    # This function relies on the fact that meson-python sets up a custom
    # loader for editable installs
    # Alternatively, one could use the distribution metadata as in:
    #  https://github.com/scientific-python/spin/blob/89e581c7201d0f6597ffc92c3e84894f99fc133b/spin/cmds/meson.py#L39
    return type(sage.__loader__).__module__ == "_sagemath_editable_loader"


def get_editable_root() -> tuple[Path, Path] | None:
    """
    Return the path to the Sage directory when using an editable
    install.
    Both the actual source directory and the build directory are returned, and are
    guaranteed to exist.
    If not using an editable install, or if the source/build directories do not
    exist, return None.

    EXAMPLES::

        sage: from sage.config import get_editable_root
        sage: get_editable_root() # random
        (WindowsPath('<path_to_sage>/sage'), WindowsPath('<path_to_sage>/sage/build/cp312'))
    """
    if (
        not is_editable_install()
        or r"/build/sagemath/src/sage" == ""
        or r"/build/sagemath/src/sage/.mesonpy-3e9q7kgd" == ""
    ):
        return None

    src = Path(r"/build/sagemath/src/sage").resolve()
    build = Path(r"/build/sagemath/src/sage/.mesonpy-3e9q7kgd").resolve()
    if src.is_dir() and build.is_dir():
        return src, build
    return None


def get_include_dirs() -> list[Path]:
    """
    Return a list of directories to be used as include directories
    when compiling Cython extensions that depend on Sage.

    Headers should be included with the prefix "sage/", e.g.,
    ``#include <sage/cpython/cython_metaclass.h>``.

    EXAMPLES::

        sage: from sage.config import get_include_dirs
        sage: dirs = get_include_dirs()
        sage: dirs # random
        [
            WindowsPath('<python>/site-packages'),
            WindowsPath('<path_to_sage>/src'),
            WindowsPath('<path_to_sage>/build/cp312/src'),
            WindowsPath('<python>/site-packages/numpy/core/include')
        ]
    """
    dirs: list[Path] = [Path(dir).parent for dir in sage.__path__]
    editable_root = get_editable_root()
    if editable_root is not None:
        # We return both the source and build directory,
        # because some headers are generated in the build directory.
        dirs.extend([root / "src" for root in editable_root])
    return [dir for dir in dirs if dir.is_dir()]



"""
Tornaria methods for computing with quadratic forms
"""
# ****************************************************************************
#       Copyright (C) 2007 Gonzalo Tornaria
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from sage.arith.misc import (CRT_vectors,
                             factor,
                             gcd,
                             hilbert_symbol,
                             kronecker as kronecker_symbol,
                             prime_to_m_part)
from sage.misc.functional import is_odd
from sage.misc.misc_c import prod
from sage.modules.free_module import FreeModule
from sage.modules.free_module_element import vector
from sage.rings.integer_ring import ZZ


# TO DO -- Add second argument
#  def __call__(self, v, w=None):
#    if w is None:
#        return half(v * self._matrix_() * v)
#    else:
#      return v * self._matrix_() * w


def disc(self):
    r"""
    Return the discriminant of the quadratic form, defined as

    - `(-1)^n {\rm det}(B)` for even dimension `2n`
    - `{\rm det}(B)/2` for odd dimension

    where `2Q(x) = x^t B x`.

    This agrees with the usual discriminant for binary and ternary quadratic forms.

    EXAMPLES::

        sage: DiagonalQuadraticForm(ZZ, [1]).disc()
        1
        sage: DiagonalQuadraticForm(ZZ, [1,1]).disc()
        -4
        sage: DiagonalQuadraticForm(ZZ, [1,1,1]).disc()
        4
        sage: DiagonalQuadraticForm(ZZ, [1,1,1,1]).disc()
        16
    """
    if is_odd(self.dim()):
        # This is not so good for characteristic 2.
        return self.base_ring()(self.det() / 2)
    return (-1)**(self.dim() // 2) * self.det()


def content(self):
    r"""
    Return the GCD of the coefficients of the quadratic form.

    .. warning::

        Only works over Euclidean domains (probably just `\ZZ`).

    EXAMPLES::

        sage: Q = DiagonalQuadraticForm(ZZ, [1, 1])
        sage: Q.matrix().gcd()
        2
        sage: Q.content()
        1
        sage: DiagonalQuadraticForm(ZZ, [1, 1]).is_primitive()
        True
        sage: DiagonalQuadraticForm(ZZ, [2, 4]).is_primitive()
        False
        sage: DiagonalQuadraticForm(ZZ, [2, 4]).primitive()
        Quadratic form in 2 variables over Integer Ring with coefficients:
        [ 1 0 ]
        [ * 2 ]
    """
    return self.gcd()


# in quadratic_form.py
# def is_primitive(self) -> bool:
#     """
#     Checks if the form is a multiple of another form... only over ZZ for now.
#     """
#     return self.content() == 1


# in quadratic_form.py
# def primitive(self):
#     """
#     Return a primitive quadratic forms in the similarity class of the given form.
#
#     This only works when we have GCDs... so over ZZ.
#     """
#     c = self.content()
#     new_coeffs = [self.base_ring()(a/c)  for a in self.__coeffs]
#     return QuadraticForm(self.base_ring(), self.dim(), new_coeffs)


def adjoint(self):
    """
    Return the adjoint (integral) quadratic form associated to the
    given form, essentially defined by taking the adjoint of the matrix.

    EXAMPLES::

        sage: Q = QuadraticForm(ZZ, 2, [1,2,5])
        sage: Q.adjoint()
        Quadratic form in 2 variables over Integer Ring with coefficients:
        [ 5 -2 ]
        [ * 1 ]

    ::

        sage: Q = QuadraticForm(ZZ, 3, [1, 0, -1, 2, -1, 5])
        sage: Q.adjoint()
        Quadratic form in 3 variables over Integer Ring with coefficients:
        [ 39 2 8 ]
        [ * 19 4 ]
        [ * * 8 ]
    """
    from sage.quadratic_forms.quadratic_form import QuadraticForm
    if is_odd(self.dim()):
        return QuadraticForm(self.matrix().adjoint_classical() * 2)
    return QuadraticForm(self.matrix().adjoint_classical())


def antiadjoint(self):
    """
    This gives an (integral) form such that its adjoint is the given form.

    EXAMPLES::

        sage: Q = QuadraticForm(ZZ, 3, [1, 0, -1, 2, -1, 5])
        sage: Q.adjoint().antiadjoint()
        Quadratic form in 3 variables over Integer Ring with coefficients:
        [ 1 0 -1 ]
        [ * 2 -1 ]
        [ * * 5 ]
        sage: Q.antiadjoint()                                                           # needs sage.symbolic
        Traceback (most recent call last):
        ...
        ValueError: not an adjoint
    """
    n = self.dim()
    R = self.base_ring()
    try:
        d = R(self.disc()**(ZZ.one() / (n - 1)))
        if is_odd(n):
            return self.adjoint().scale_by_factor(R.one() / 4 / d**(n - 2))
        return self.adjoint().scale_by_factor(R.one() / d**(n - 2))
    except TypeError:
        raise ValueError("not an adjoint")


def is_adjoint(self) -> bool:
    """
    Determine if the given form is the adjoint of another form.

    EXAMPLES::

        sage: Q = QuadraticForm(ZZ, 3, [1, 0, -1, 2, -1, 5])
        sage: Q.is_adjoint()                                                            # needs sage.symbolic
        False
        sage: Q.adjoint().is_adjoint()
        True
    """
    try:
        self.antiadjoint()
    except ValueError:
        return False
    return True


def reciprocal(self):
    r"""
    This gives the reciprocal quadratic form associated to the given form.

    This is defined as the multiple of the primitive adjoint with the same
    content as the given form.

    EXAMPLES::

        sage: Q = DiagonalQuadraticForm(ZZ, [1,1,37])
        sage: Q.reciprocal()
        Quadratic form in 3 variables over Integer Ring with coefficients:
        [ 37 0 0 ]
        [ * 37 0 ]
        [ * * 1 ]
        sage: Q.reciprocal().reciprocal()
        Quadratic form in 3 variables over Integer Ring with coefficients:
        [ 1 0 0 ]
        [ * 1 0 ]
        [ * * 37 ]
        sage: Q.reciprocal().reciprocal() == Q
        True
    """
    return self.adjoint().primitive() . scale_by_factor(self.content())


def omega(self):
    r"""
    Return the content of the adjoint of the primitive associated quadratic form.

    Ref: See Dickson's "Studies in Number Theory".

    EXAMPLES::

        sage: Q = DiagonalQuadraticForm(ZZ, [1,1,37])
        sage: Q.omega()
        4
    """
    return self.primitive().adjoint().content()


def delta(self):
    r"""
    Return the omega of the adjoint form.

    This is the same as the omega of the reciprocal form.

    EXAMPLES::

        sage: Q = DiagonalQuadraticForm(ZZ, [1,1,37])
        sage: Q.delta()
        148
    """
    return self.adjoint().omega()


def level__Tornaria(self):
    r"""
    Return the level of the quadratic form.

    This is defined as

    - level(`B`)    for even dimension,
    - level(`B`)/4  for odd dimension,

    where `2Q(x) = x^t\cdot B\cdot x`.

    This agrees with the usual level for even dimension.

    EXAMPLES::

        sage: DiagonalQuadraticForm(ZZ, [1]).level__Tornaria()
        1
        sage: DiagonalQuadraticForm(ZZ, [1,1]).level__Tornaria()
        4
        sage: DiagonalQuadraticForm(ZZ, [1,1,1]).level__Tornaria()
        1
        sage: DiagonalQuadraticForm(ZZ, [1,1,1,1]).level__Tornaria()
        4
    """
    return self.base_ring()(abs(self.disc()) / self.omega() / self.content()**self.dim())


def discrec(self):
    r"""
    Return the discriminant of the reciprocal form.

    EXAMPLES::

        sage: Q = DiagonalQuadraticForm(ZZ, [1,1,37])
        sage: Q.disc()
        148
        sage: Q.discrec()
        5476
        sage: [4 * 37, 4 * 37^2]
        [148, 5476]
    """
    return self.reciprocal().disc()


# Rational equivalence

def hasse_conductor(self):
    """
    Return the Hasse conductor.

    This is the product of all primes where the Hasse invariant equals `-1`.

    EXAMPLES::

        sage: # needs sage.libs.pari
        sage: Q = QuadraticForm(ZZ, 3, [1, 0, -1, 2, -1, 5])
        sage: Q.hasse_invariant(2)
        -1
        sage: Q.hasse_invariant(37)
        -1
        sage: Q.hasse_conductor()
        74

        sage: DiagonalQuadraticForm(ZZ, [1, 1, 1]).hasse_conductor()                    # needs sage.libs.pari
        1
        sage: QuadraticForm(ZZ, 3, [2, -2, 0, 2, 0, 5]).hasse_conductor()               # needs sage.libs.pari
        10
    """
    return prod([x for x, _ in factor(2 * self.level())
                 if self.hasse_invariant(x) == -1])


def clifford_invariant(self, p):
    """
    Return the Clifford invariant.

    This is the class in the Brauer group of the Clifford algebra for
    even dimension, of the even Clifford Algebra for odd dimension.

    See Lam (AMS GSM 67) p. 117 for the definition, and p. 119 for the formula
    relating it to the Hasse invariant.

    EXAMPLES:

    For hyperbolic spaces, the Clifford invariant is +1::

        sage: # needs sage.libs.pari
        sage: H = QuadraticForm(ZZ, 2, [0, 1, 0])
        sage: H.clifford_invariant(2)
        1
        sage: (H + H).clifford_invariant(2)
        1
        sage: (H + H + H).clifford_invariant(2)
        1
        sage: (H + H + H + H).clifford_invariant(2)
        1
    """
    n = self.dim() % 8
    if n == 1 or n == 2:
        s = 1
    elif n == 3 or n == 4:
        s = hilbert_symbol(-1, -self.disc(), p)
    elif n == 5 or n == 6:
        s = hilbert_symbol(-1, -1, p)
    elif n == 7 or n == 0:
        s = hilbert_symbol(-1, self.disc(), p)
    return s * self.hasse_invariant(p)


def clifford_conductor(self):
    """
    Return the product of all primes where the Clifford invariant is `-1`.

    .. NOTE::

        For ternary forms, this is the discriminant of the
        quaternion algebra associated to the quadratic space
        (i.e. the even Clifford algebra).

    EXAMPLES::

        sage: # needs sage.libs.pari
        sage: Q = QuadraticForm(ZZ, 3, [1, 0, -1, 2, -1, 5])
        sage: Q.clifford_invariant(2)
        1
        sage: Q.clifford_invariant(37)
        -1
        sage: Q.clifford_conductor()
        37

        sage: DiagonalQuadraticForm(ZZ, [1, 1, 1]).clifford_conductor()                 # needs sage.libs.pari
        2
        sage: QuadraticForm(ZZ, 3, [2, -2, 0, 2, 0, 5]).clifford_conductor()            # needs sage.libs.pari
        30

    For hyperbolic spaces, the Clifford conductor is 1::

        sage: # needs sage.libs.pari
        sage: H = QuadraticForm(ZZ, 2, [0, 1, 0])
        sage: H.clifford_conductor()
        1
        sage: (H + H).clifford_conductor()
        1
        sage: (H + H + H).clifford_conductor()
        1
        sage: (H + H + H + H).clifford_conductor()
        1
    """
    return prod([x for x, _ in factor(2 * self.level())
                 if self.clifford_invariant(x) == -1])


# Genus theory

def basiclemma(self, M):
    """
    Find a number represented by ``self`` and coprime to `M`.

    EXAMPLES::

        sage: Q = QuadraticForm(ZZ, 2, [2, 1, 3])
        sage: Q.basiclemma(6)
        71
    """
    a = self(self.basiclemmavec(M))
    assert gcd(a, M) == 1
    return a


def basiclemmavec(self, M):
    """
    Find a vector where the value of the quadratic form is coprime to `M`.

    EXAMPLES::

        sage: Q = QuadraticForm(ZZ, 2, [2, 1, 5])
        sage: Q.basiclemmavec(10)
        (6, 5)
        sage: Q(_)
        227
    """
    V = FreeModule(self.base_ring(), self.dim())
    vec = []
    mod = []
    M0 = abs(M)
    if M0 == 1:
        return V.zero()

    for i in range(self.dim()):
        M1 = prime_to_m_part(M0, self[i, i])
        if M1 != 1:
            vec.append(V.gen(i))
            mod.append(M1)
        M0 = M0 / M1
        if M0 == 1:
            return tuple(CRT_vectors(vec, mod))

    for i in range(self.dim()):
        for j in range(i):
            M1 = prime_to_m_part(M0, self[i, j])
            if M1 != 1:
                vec.append(V.i + V.j)
                mod.append(M1)
            M0 = M0 / M1
            if M0 == 1:
                return tuple(CRT_vectors(vec, mod))

    raise ValueError("not primitive form")


# FIXME: get the rules for validity of characters straight...
# p=2 might be bad!!!
def xi(self, p):
    """
    Return the value of the genus characters Xi_p... which may be missing one character.
    We allow -1 as a prime.

    REFERENCES:

    Dickson's "Studies in the Theory of Numbers"

    EXAMPLES::

        sage: Q1 = QuadraticForm(ZZ, 3, [1, 1, 1, 14, 3, 14])
        sage: Q2 = QuadraticForm(ZZ, 3, [2, -1, 0, 2, 0, 50])
        sage: [Q1.omega(), Q2.omega()]
        [5, 5]
        sage: [Q1.hasse_invariant(5),                   # equivalent over Q_5           # needs sage.libs.pari
        ....:  Q2.hasse_invariant(5)]
        [1, 1]
        sage: [Q1.xi(5), Q2.xi(5)]                      # not equivalent over Z_5       # needs sage.libs.pari
        [1, -1]
    """
    if self.dim() == 2 and self.disc() % p:
        raise ValueError("not a valid character")
    if self.dim() >= 3 and self.omega() % p:
        raise ValueError("not a valid character")
    if (p == -1) or (p == 2):
        return kronecker_symbol(p, self.basiclemma(2))
    return kronecker_symbol(self.basiclemma(p), p)


def xi_rec(self, p):
    """
    Return Xi(`p`) for the reciprocal form.

    EXAMPLES::

        sage: # needs sage.libs.pari
        sage: Q1 = QuadraticForm(ZZ, 3, [1, 1, 1, 14, 3, 14])
        sage: Q2 = QuadraticForm(ZZ, 3, [2, -1, 0, 2, 0, 50])
        sage: [Q1.clifford_conductor(),                 # equivalent over Q
        ....:  Q2.clifford_conductor()]
        [3, 3]
        sage: Q1.is_locally_equivalent_to(Q2)           # not in the same genus
        False
        sage: [Q1.delta(), Q2.delta()]
        [480, 480]
        sage: factor(480)
        2^5 * 3 * 5
        sage: list(map(Q1.xi_rec, [-1,2,3,5]))
        [-1, -1, -1, 1]
        sage: list(map(Q2.xi_rec, [-1,2,3,5]))
        [-1, -1, -1, -1]
    """
    return self.reciprocal().xi(p)


def lll(self):
    """
    Return an LLL-reduced form of `Q` (using PARI).

    EXAMPLES::

        sage: Q = QuadraticForm(ZZ, 4, range(1,11))
        sage: Q.is_definite()
        True
        sage: Q.lll()                                                                   # needs sage.libs.pari
        Quadratic form in 4 variables over Integer Ring with coefficients:
        [ 1 0 1 0 ]
        [ * 4 3 3 ]
        [ * * 6 3 ]
        [ * * * 6 ]
    """
    return self(self.matrix().LLL_gram())


def representation_number_list(self, B):
    """
    Return the vector of representation numbers < B.

    EXAMPLES::

        sage: Q = DiagonalQuadraticForm(ZZ, [1,1,1,1,1,1,1,1])
        sage: Q.representation_number_list(10)                                          # needs sage.libs.pari
        [1, 16, 112, 448, 1136, 2016, 3136, 5504, 9328, 12112]
    """
    from sage.libs.pari import pari

    ans = pari(1).concat(self.__pari__().qfrep(B - 1, 1) * 2)
    return ans.sage()


def representation_vector_list(self, B, maxvectors=10**8):
    """
    Find all vectors `v` where `Q(v) < B`.

    This only works for positive definite quadratic forms.

    EXAMPLES::

        sage: # needs sage.libs.pari
        sage: Q = DiagonalQuadraticForm(ZZ, [1, 1])
        sage: Q.representation_vector_list(10)
        [[(0, 0)],
         [(0, 1), (0, -1), (1, 0), (-1, 0)],
         [(1, 1), (-1, -1), (1, -1), (-1, 1)],
         [],
         [(0, 2), (0, -2), (2, 0), (-2, 0)],
         [(1, 2), (-1, -2), (1, -2), (-1, 2), (2, 1), (-2, -1), (2, -1), (-2, 1)],
         [],
         [],
         [(2, 2), (-2, -2), (2, -2), (-2, 2)],
         [(0, 3), (0, -3), (3, 0), (-3, 0)]]
        sage: list(map(len, _))
        [1, 4, 4, 0, 4, 8, 0, 0, 4, 4]
        sage: Q.representation_number_list(10)
        [1, 4, 4, 0, 4, 8, 0, 0, 4, 4]

    TESTS::

        sage: R = QuadraticForm(ZZ, 2, [-4,-3,0])
        sage: R.representation_vector_list(10)                                          # needs sage.libs.pari
        Traceback (most recent call last):
        ...
        PariError: domain error in minim0: form is not positive definite
    """
    n, _, vs = self.__pari__().qfminim(2 * (B - 1), maxvectors)

    if n != 2 * len(vs):
        raise RuntimeError("insufficient number of vectors")
    ms = [[] for _ in range(B)]
    ms[0] = [vector([0] * self.dim())]
    for v in vs.sage().columns():
        ms[int(self(v))] += [v, -v]
    return ms


# zeros

def is_zero(self, v, p=0) -> bool:
    r"""
    Determine if the vector `v` is on the conic `Q(x) = 0` (mod `p`).

    EXAMPLES::

        sage: Q1 = QuadraticForm(ZZ, 3, [1, 0, -1, 2, -1, 5])
        sage: Q1.is_zero([0,1,0], 2)
        True
        sage: Q1.is_zero([1,1,1], 2)
        True
        sage: Q1.is_zero([1,1,0], 2)
        False
    """
    norm = self(v)
    if p != 0:
        norm = norm % p
    return norm == 0


def is_zero_nonsingular(self, v, p=0) -> bool:
    """
    Determine if the vector `v` is on the conic `Q(x) = 0` (mod `p`),
    and that this point is non-singular point of the conic.

    EXAMPLES::

        sage: Q1 = QuadraticForm(ZZ, 3, [1, 0, -1, 2, -1, 5])
        sage: Q1.is_zero_nonsingular([1,1,1], 2)
        True
        sage: Q1.is_zero([1, 19, 2], 37)
        True
        sage: Q1.is_zero_nonsingular([1, 19, 2], 37)
        False
    """
    if not self.is_zero(v, p):
        return False
    vm = vector(self.base_ring(), v) * self.matrix()
    if p != 0:
        vm = vm % p
    return (vm != 0)


def is_zero_singular(self, v, p=0) -> bool:
    """
    Determine if the vector `v` is on the conic `Q(x) = 0` (mod `p`),
    and that this point is singular point of the conic.

    EXAMPLES::

        sage: Q1 = QuadraticForm(ZZ, 3, [1, 0, -1, 2, -1, 5])
        sage: Q1.is_zero([1,1,1], 2)
        True
        sage: Q1.is_zero_singular([1,1,1], 2)
        False
        sage: Q1.is_zero_singular([1, 19, 2], 37)
        True
    """
    if not self.is_zero(v, p):
        return False
    vm = vector(self.base_ring(), v) * self.matrix()
    if p != 0:
        vm = vm % p
    return bool(vm == 0)



r"""
Introduction to Drinfeld modular forms
======================================

This tutorial outlines the definitions, the notations, and the
implementation of Drinfeld modular forms in SageMath. We assume that the
reader has basic knowledge of classical modular forms, as we will often
make analogies to this setting. We also assume little knowledge of
Drinfeld modules; for this topic, the interested reader can consult the
SageMath reference manual
:ref:`Drinfeld modules <sage.rings.function_field.drinfeld_modules.drinfeld_module>`.

.. RUBRIC:: Preliminary notations

Let `q` be a prime power and let `A` be the ring of functions of
`\mathbb{P}^1/\mathbb{F}_q` which are regular outside a closed point
`\infty`. This ring is the polynomial ring `\mathbb{F}_q[T]`. We denote
by `K := \mathbb{F}_q(T)` rational function field. We endow `K` with the
`1/T`-adic valuation and let `K_{\infty} := \mathbb{F}_q((1/T))` be the
completion of `K`. Next, we define `\mathbb{C}_{\infty}` to be the
completion of an algebraic closure of `K_{\infty}`. Lastly, we denote
by `\tau : x\mapsto x^q` the `q`-Frobenius.

.. NOTE::

    The above construction of `\mathbb{C}_{\infty}` is the same as the
    construction of `\mathbb{C}_p` in the case of `p`-adic numbers
    (see :wikipedia:`P-adic_number#Algebraic_closure`).

In SageMath, we create the rational function field by first creating a
univariate polynomial ring over `\mathbb{F}_q` and, following this, by
constructing its field of fractions::

    sage: A = GF(3)['T']
    sage: K.<T> = Frac(A)
    sage: K
    Fraction Field of Univariate Polynomial Ring in T over Finite Field of size 3
    sage: K.base()  # returns A
    Univariate Polynomial Ring in T over Finite Field of size 3

.. RUBRIC:: Drinfeld period domain and action of `\mathrm{GL}_r(K_{\infty})`

In the classical setting, the domain of any modular form is the complex
upper half plane `\mathcal{H}:=\{w\in \mathbb{C} : \mathrm{im}(w)>0\}`.
The analogue of this plane in the function field setting is the
*Drinfeld period domain of rank* `r > 1` and it is defined by

.. MATH::

    \Omega^r(\mathbb{C}_{\infty}) :=
    \mathbb{P}^{r-1}(\mathbb{C}_{\infty})
    \setminus \{K_{\infty}\text{-rational hyperplanes}\}.

This space is a rigid analytic space and, after fixing an arbitrary
nonzero constant `\xi` in `\mathbb{C}_{\infty}`, we identify its
elements with the set of column vectors
`(w_1,\ldots, w_{r-1}, w_{r})^{\mathrm{T}}` in `\mathbb{C}_{\infty}^r`
such that the `w_i` are `K_{\infty}`-linearly independant and
`w_r = \xi`. Note that `\xi` is unspecified, but the reader can assume
that `\xi = 1` without any loss of significant information. Its value
can be interesting simply for normalization purposes.

We define a left action of `\mathrm{GL}_r(K_{\infty})` on
`\Omega^r(\mathbb{C}_{\infty})` by setting

.. MATH::

    \gamma(w) := j(\gamma, w)^{-1}\gamma w

where `j(\gamma, w) := \xi^{-1} \cdot (\text{last entry of }\gamma w)`.

.. RUBRIC:: Universal Drinfeld module over `\Omega^r(\mathbb{C}_{\infty})`

For any `w = (w_1, \ldots, w_{r-1}, \xi)` in
`\Omega^r(\mathbb{C}_{\infty})` we have a corresponding discrete
`A`-module `\Lambda^w` which is free of rank `r`:

.. MATH::

    \Lambda^w := Aw_1 \oplus \cdots \oplus Aw_{r-1} \oplus A\xi.

An important result is that we have analytic uniformization which is the
analogue of complex uniformization for elliptic curves. In our setting,
elliptic curves are replaced by Drinfeld modules. In short, there exists
a corresponding Drinfeld module

.. MATH::

    \phi^w : T \mapsto T + g_1(w)\tau + \cdots
    + g_{r - 1}(w)\tau^{r-1} + g_{r}(w)\tau^{r}.

such that the exponential of `\phi^w` induces an isomorphism (of abelian
group) between the additive group `\mathbb{C}_{\infty}` and the quotient
`\mathbb{C}_{\infty} / \Lambda^w`. Background material on Drinfeld
modules and their analytic uniformization can be found in section 4.3
and 4.6 of [Gos1998]_.

The Drinfeld module
`\phi^w : A \to \mathbb{C}_{\infty} \{\tau\}` is called the
*universal Drinfeld* `\mathbb{F}_q[T]`-*module over*
`\Omega^r(\mathbb{C}_{\infty})` and its coefficients
`g_i : \Omega^r(\mathbb{C}_{\infty}) \to \mathbb{C}_{\infty}`
are rigid analytic functions satisfying the *invariance property*:

.. MATH::

    g_i(\gamma(w)) = j(\gamma, w)^{1 - q^i} g_i(w),
    ~\forall \gamma\in \mathrm{GL}_r(A)

where `g_{r}(w)` never vanishes. Moreover, these coefficients `g_i`
admits an expansion at infinity, analogous to `q`-expansion principle
for classical modular forms. The functions `g_i` are known as the
*coefficients forms at* `T`. More generally, the coefficients of the
image `\phi^w_a` for any `a\in A` are called the *coefficient forms at*
`a` and they are an algebraic combination of the coefficient forms at
`T`.

In the rank two case, the expansion at infinity is of the form

.. MATH::

    g_i(w) = \sum_{i = 0}^{\infty} a_n(g_i)u(w)^i

where `u(w) := e(w)^{-1}` and `e` is the exponential function of the
Carlitz module `\rho:T\mapsto T + \tau`. The analytic parameter `u` is
called the *parameter at infinity*.

A *Drinfeld modular form* of rank `r`, weight `k`, type `m` for
`\mathrm{GL}_r(A)` is a rigid analytic function

.. MATH::

    f:\Omega^r(\mathbb{C}_{\infty}) \to \mathbb{C}_{\infty}

such that

* `f(\gamma(w)) = \mathrm{det}(\gamma)^m j(\gamma, w)^k f(w)` for all
  `\gamma` in `\mathrm{GL}_r(A)` and
  `w\in \Omega^r(\mathbb{C}_{\infty})`;

* `f` is holomorphic at infinity.

Without diving into the details, we mention that the second condition is
similar to the classical case. More specifically, in the rank two
situation, the expansion of `f` is given by a power series in `u`
`f = \sum_{n\geq 0} a_n(f) u^n` where `a_n(f)\in \mathbb{C}_{\infty}`.

Lastly, we also mention that the integer `m` only depends on its class
modulo `q-1`.

Note that all the above theory is covered in much greater details in
part I of [BRP2018]_.

.. RUBRIC:: Ring of Drinfeld modular forms

Letting `M_k^{r, m}(\mathrm{GL}_r(A))` denote the space of rank `r`,
weight `k\in (q - 1)\mathbb{Z}` and type `m` Drinfeld modular forms,
we define

.. MATH::

    M^{r, 0}(\mathrm{GL}_r(A)) :=
    \bigoplus_{k\in ZZ} M_k^{r, 0}(\mathrm{GL}_r(A))

to be the graded ring of all Drinfeld modular forms of type 0. The
graduation is given by the weight of a modular form. Similarly, we let
`M^{r}(\mathrm{GL}_r(A)) \supset M^{r, 0}(\mathrm{GL}_r(A))` be the ring
of all Drinfeld modular forms of rank `r` and arbitrary type. By
theorem 17.5 in part III of [BRP2018]_, we have

.. MATH::

    M^{r, 0}(\mathrm{GL}_r(A))
    = \mathbb{C}_{\infty}[g_1,\ldots, g_{r-1}, g_{r}].

and

.. MATH::

    M^{r}(\mathrm{GL}_r(A))
    = \mathbb{C}_{\infty}[g_1,\ldots, g_{r-1}, h_{r}].

where `h_r` is a weight `(q^r - 1)/(q - 1)` modular forms of type `1`
which is a `(q-1)`-root of `g_r` sometimes known as *Gekeler's* `h`
*function*, see theorem 3.8 of [Gek2017]_ for the precise definition of
this function.

.. RUBRIC:: SageMath implementation

In SageMath, we model the ring of type 0 Drinfeld modular forms over `K`
as a finitely generated ring in the coefficients forms `g_i`:

.. MATH::

    K[g_1,\ldots, g_{r-1}, g_r].

Hence, any ring element is seen as a formal algebraic combination of the
coefficient forms `g_i` over `K`. Likewise, the ring of arbitrary
type forms is generated by `g_1\ldots, g_{r-1}, h_r`.

To create the ring of type zero and rank `r` Drinfeld modular forms, one
uses the class
:class:`~sage.modular.drinfeld_modform.ring.DrinfeldModularForms`::

    sage: A = GF(3)['T']
    sage: K.<T> = Frac(A)
    sage: M = DrinfeldModularForms(K, 3)  # rank 3
    sage: M
    Ring of Drinfeld modular forms of rank 3 over Fraction Field of Univariate Polynomial Ring in T over Finite Field of size 3

To create the ring of arbitrary types modular forms, one passes the
keyword argument ``has_type=True``::

    sage: M = DrinfeldModularForms(K, 4, has_type=True)
    sage: M.gens()
    (g1, g2, g3, h4)
    sage: h4 = M.3
    sage: h4.weight()
    40

For more information about the functionalities of the implementation,
one should consult the documentation of the main classes:

- Parent class:
  :class:`~sage.modular.drinfeld_modform.ring.DrinfeldModularForms`

- Element class:
  :class:`~sage.modular.drinfeld_modform.element.DrinfeldModularFormsElement`

.. RUBRIC:: References

A good introduction to Drinfeld modular forms of rank 2, see Gekeler's
paper [Gek1988]_. See also [BRP2018]_ for a detailed exposition of the
arbitrary rank theory.
"""



"""
Disks
"""
# *****************************************************************************
#       Copyright (C) 2006 Alex Clemesha <clemesha@gmail.com>,
#                          William Stein <wstein@gmail.com>,
#                     2008 Mike Hansen <mhansen@gmail.com>,
#                     2021 Javier Honrubia <jhonrubia6@alumno.uned.es>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  http://www.gnu.org/licenses/
# *****************************************************************************

from sage.plot.primitive import GraphicPrimitive
from sage.misc.decorators import options, rename_keyword
from sage.plot.colors import to_mpl_color
from math import sin, cos, pi


class Disk(GraphicPrimitive):
    """
    Primitive class for the ``Disk`` graphics type.  See ``disk?`` for
    information about actually plotting a disk (the Sage term for a sector
    or wedge of a circle).

    INPUT:

    - ``point`` -- coordinates of center of disk

    - ``r`` -- radius of disk

    - ``angle`` -- beginning and ending angles of disk (i.e.
      angle extent of sector/wedge)

    - ``options`` -- dictionary of valid plot options to pass to constructor

    EXAMPLES:

    Note this should normally be used indirectly via ``disk``::

        sage: from math import pi
        sage: from sage.plot.disk import Disk
        sage: D = Disk((1,2), 2, (pi/2,pi), {'zorder':3})
        sage: D
        Disk defined by (1.0,2.0) with r=2.0
         spanning (1.5707963267..., 3.1415926535...) radians
        sage: D.options()['zorder']
        3
        sage: D.x
        1.0

    TESTS:

    We test creating a disk::

        sage: disk((2,3), 2, (0,pi/2))
        Graphics object consisting of 1 graphics primitive
    """
    def __init__(self, point, r, angle, options):
        """
        Initialize base class ``Disk``.

        EXAMPLES::

            sage: from math import pi
            sage: D = disk((2,3), 1, (pi/2, pi), fill=False, color='red', thickness=1, alpha=.5)
            sage: D[0].x
            2.0
            sage: D[0].r
            1.0
            sage: D[0].rad1
            1.5707963267948966
            sage: D[0].options()['rgbcolor']
            'red'
            sage: D[0].options()['alpha']
            0.500000000000000
            sage: print(loads(dumps(D)))
            Graphics object consisting of 1 graphics primitive
        """
        self.x = float(point[0])
        self.y = float(point[1])
        self.r = float(r)
        self.rad1 = float(angle[0])
        self.rad2 = float(angle[1])
        GraphicPrimitive.__init__(self, options)

    def get_minmax_data(self):
        """
        Return a dictionary with the bounding box data.

        EXAMPLES::

            sage: from math import pi
            sage: D = disk((5,4), 1, (pi/2, pi))
            sage: d = D.get_minmax_data()
            sage: d['xmin']
            4.0
            sage: d['ymin']
            3.0
            sage: d['xmax']
            6.0
            sage: d['ymax']
            5.0
        """
        from sage.plot.plot import minmax_data
        return minmax_data([self.x - self.r, self.x + self.r],
                           [self.y - self.r, self.y + self.r],
                           dict=True)

    def _allowed_options(self):
        """
        Return the allowed options for the ``Disk`` class.

        EXAMPLES::

            sage: from math import pi
            sage: p = disk((3, 3), 1, (0, pi/2))
            sage: p[0]._allowed_options()['alpha']
            'How transparent the figure is.'
            sage: p[0]._allowed_options()['zorder']
            'The layer level in which to draw'
        """
        return {'alpha': 'How transparent the figure is.',
                'fill': 'Whether or not to fill the disk.',
                'legend_label': 'The label for this item in the legend.',
                'legend_color': 'The color of the legend text.',
                'thickness': 'How thick the border of the disk is.',
                'rgbcolor': 'The color as an RGB tuple.',
                'hue': 'The color given as a hue.',
                'zorder': 'The layer level in which to draw'}

    def _repr_(self):
        """
        String representation of ``Disk`` primitive.

        EXAMPLES::

            sage: from math import pi
            sage: P = disk((3, 3), 1, (0, pi/2))
            sage: p = P[0]; p
            Disk defined by (3.0,3.0) with r=1.0 spanning (0.0, 1.5707963267...) radians
        """
        return "Disk defined by ({},{}) with r={} spanning ({}, {}) radians".format(self.x, self.y, self.r, self.rad1, self.rad2)

    def _render_on_subplot(self, subplot):
        """
        TESTS::

            sage: from math import pi
            sage: D = disk((2,-1), 2, (0, pi), color='black', thickness=3, fill=False); D
            Graphics object consisting of 1 graphics primitive

        Save alpha information in pdf (see :issue:`13732`)::

            sage: f = tmp_filename(ext='.pdf')
            sage: p = disk((0,0), 5, (0, pi/4), alpha=0.5)
            sage: p.save(f)
        """
        from matplotlib import patches
        options = self.options()
        deg1 = self.rad1*(180./pi)  # convert radians to degrees
        deg2 = self.rad2*(180./pi)
        z = int(options.pop('zorder', 0))
        p = patches.Wedge((float(self.x), float(self.y)), float(self.r), float(deg1),
                          float(deg2), zorder=z)
        a = float(options['alpha'])
        p.set_alpha(a)
        p.set_linewidth(float(options['thickness']))
        p.set_fill(options['fill'])
        c = to_mpl_color(options['rgbcolor'])
        p.set_edgecolor(c)
        p.set_facecolor(c)
        p.set_label(options['legend_label'])
        subplot.add_patch(p)

    def plot3d(self, z=0, **kwds):
        """
        Plots a 2D disk (actually a 52-gon) in 3D,
        with default height zero.

        INPUT:

        - ``z`` -- (optional) 3D height above `xy`-plane

        AUTHORS:

        - Karl-Dieter Crisman (05-09)

        EXAMPLES::

            sage: from math import pi
            sage: disk((0,0), 1, (0, pi/2)).plot3d()
            Graphics3d Object
            sage: disk((0,0), 1, (0, pi/2)).plot3d(z=2)
            Graphics3d Object
            sage: disk((0,0), 1, (pi/2, 0), fill=False).plot3d(3)
            Graphics3d Object


        These examples show that the appropriate options are passed::

            sage: D = disk((2,3), 1, (pi/4,pi/3), hue=.8, alpha=.3, fill=True)
            sage: d = D[0]
            sage: d.plot3d(z=2).texture.opacity
            0.3

        ::

            sage: D = disk((2,3), 1, (pi/4,pi/3), hue=.8, alpha=.3, fill=False)
            sage: d = D[0]
            sage: dd = d.plot3d(z=2)
            sage: dd.jmol_repr(dd.testing_render_params())[0][-1]
            'color $line_4 translucent 0.7 [204,0,255]'
        """
        options = dict(self.options())
        fill = options['fill']
        del options['fill']
        if 'zorder' in options:
            del options['zorder']
        n = 50
        x, y, r, rad1, rad2 = self.x, self.y, self.r, self.rad1, self.rad2
        dt = float((rad2-rad1)/n)
        xdata = [x]
        ydata = [y]
        xdata.extend([x+r*cos(t*dt+rad1) for t in range(n+1)])
        ydata.extend([y+r*sin(t*dt+rad1) for t in range(n+1)])
        xdata.append(x)
        ydata.append(y)
        if fill:
            from .polygon import Polygon
            return Polygon(xdata, ydata, options).plot3d(z)
        from .line import Line
        return Line(xdata, ydata, options).plot3d().translate((0, 0, z))


@rename_keyword(color='rgbcolor')
@options(alpha=1, fill=True, rgbcolor=(0, 0, 1), thickness=0, legend_label=None,
         legend_color=None, aspect_ratio=1.0)
def disk(point, radius, angle, **options):
    r"""
    A disk (that is, a sector or wedge of a circle) with center
    at a point = `(x,y)` (or `(x,y,z)` and parallel to the
    `xy`-plane) with radius = `r` spanning (in radians)
    angle=`(rad1, rad2)`.

    Type ``disk.options`` to see all options.

    EXAMPLES:

    Make some dangerous disks::

        sage: from math import pi
        sage: bl = disk((0.0,0.0), 1, (pi, 3*pi/2), color='yellow')
        sage: tr = disk((0.0,0.0), 1, (0, pi/2), color='yellow')
        sage: tl = disk((0.0,0.0), 1, (pi/2, pi), color='black')
        sage: br = disk((0.0,0.0), 1, (3*pi/2, 2*pi), color='black')
        sage: P  = tl + tr + bl + br
        sage: P.show(xmin=-2, xmax=2, ymin=-2, ymax=2)

    .. PLOT::

        from sage.plot.disk import Disk
        bl = disk((0.0,0.0), 1, (pi, 3*pi/2), color='yellow')
        tr = disk((0.0,0.0), 1, (0, pi/2), color='yellow')
        tl = disk((0.0,0.0), 1, (pi/2, pi), color='black')
        br = disk((0.0,0.0), 1, (3*pi/2, 2*pi), color='black')
        P  = tl+tr+bl+br
        sphinx_plot(P)

    The default aspect ratio is 1.0::

        sage: disk((0.0,0.0), 1, (pi, 3*pi/2)).aspect_ratio()
        1.0

    Another example of a disk::

        sage: bl = disk((0.0,0.0), 1, (pi, 3*pi/2), rgbcolor=(1,1,0))
        sage: bl.show(figsize=[5,5])

    .. PLOT::

        from sage.plot.disk import Disk
        bl = disk((0.0,0.0), 1, (pi, 3*pi/2), rgbcolor=(1,1,0))
        sphinx_plot(bl)

    Note that since ``thickness`` defaults to zero, it is best to change
    that option when using ``fill=False``::

        sage: disk((2,3), 1, (pi/4,pi/3), hue=.8, alpha=.3, fill=False, thickness=2)
        Graphics object consisting of 1 graphics primitive

    .. PLOT::

        from sage.plot.disk import Disk
        D = disk((2,3), 1, (pi/4,pi/3), hue=.8, alpha=.3, fill=False, thickness=2)
        sphinx_plot(D)

    The previous two examples also illustrate using ``hue`` and ``rgbcolor``
    as ways of specifying the color of the graphic.

    We can also use this command to plot three-dimensional disks parallel
    to the `xy`-plane::

        sage: d = disk((1,1,3), 1, (pi,3*pi/2), rgbcolor=(1,0,0))
        sage: d
        Graphics3d Object
        sage: type(d)
        <... 'sage.plot.plot3d.index_face_set.IndexFaceSet'>

    .. PLOT::

        from sage.plot.disk import Disk
        d = disk((1,1,3), 1, (pi,3*pi/2), rgbcolor=(1,0,0))
        sphinx_plot(d)

    Extra options will get passed on to ``show()``, as long as they are valid::

        sage: disk((0, 0), 5, (0, pi/2), rgbcolor=(1, 0, 1),
        ....:      xmin=0, xmax=5, ymin=0, ymax=5, figsize=(2,2))
        Graphics object consisting of 1 graphics primitive
        sage: disk((0, 0), 5, (0, pi/2), rgbcolor=(1, 0, 1)).show(  # These are equivalent
        ....:     xmin=0, xmax=5, ymin=0, ymax=5, figsize=(2,2))

    TESTS:

    Testing that legend labels work right::

        sage: disk((2,4), 3, (pi/8, pi/4), hue=1, legend_label='disk', legend_color='blue')
        Graphics object consisting of 1 graphics primitive

    We cannot currently plot disks in more than three dimensions::

        sage: d = disk((1,1,1,1), 1, (0,pi))
        Traceback (most recent call last):
        ...
        ValueError: the center point of a plotted disk should have two or three coordinates

    Verify that :issue:`36153` is fixed::

        sage: D = disk((0, 0), 5, (0, pi/2), legend_label='test')
    """
    from sage.plot.graphics import Graphics
    g = Graphics()

    # Reset aspect_ratio to 'automatic' in case scale is 'semilog[xy]'.
    # Otherwise matplotlib complains.
    scale = options.get('scale', None)
    if isinstance(scale, (list, tuple)):
        scale = scale[0]
    if scale == 'semilogy' or scale == 'semilogx':
        options['aspect_ratio'] = 'automatic'

    g._set_extra_kwds(Graphics._extract_kwds_for_show(options))
    g.add_primitive(Disk(point, radius, angle, options))
    if options['legend_label']:
        g.legend(True)
        g._legend_colors = [options['legend_color']]
    if len(point) == 2:
        return g
    if len(point) == 3:
        return g[0].plot3d(z=point[2])
    raise ValueError('the center point of a plotted disk should have '
                     'two or three coordinates')



r"""
Homomorphisms of finitely generated free graded left modules

AUTHORS:

- Robert R. Bruner, Michael J. Catanzaro (2012): Initial version.
- Sverre Lunoee--Nielsen and Koen van Woerden (2019-11-29): Updated the
  original software to Sage version 8.9.
- Sverre Lunoee--Nielsen (2020-07-01): Refactored the code and added
  new documentation and tests.
"""

# *****************************************************************************
#       Copyright (C) 2019 Robert R. Bruner <rrb@math.wayne.edu>
#                     and  Michael J. Catanzaro <mike@math.wayne.edu>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from sage.modules.fp_graded.morphism import FPModuleMorphism


class FreeGradedModuleMorphism(FPModuleMorphism):
    r"""
    Create a homomorphism from a finitely generated free graded module
    to a graded module.

    INPUT:

    - ``parent`` -- a homspace in the category of finitely generated free
      modules

    - ``values`` -- list of elements in the codomain; each element
      corresponds (by their ordering) to a module generator in the domain

    EXAMPLES::

        sage: from sage.modules.fp_graded.free_module import FreeGradedModule
        sage: A = SteenrodAlgebra(2)
        sage: F1 = FreeGradedModule(A, (4,5), names='b')
        sage: F2 = FreeGradedModule(A, (3,4), names='c')
        sage: F3 = FreeGradedModule(A, (2,3), names='d')
        sage: H1 = Hom(F1, F2)
        sage: H2 = Hom(F2, F3)
        sage: f = H1((F2((Sq(4), 0)), F2((0, Sq(4)))))
        sage: g = H2((F3((Sq(2), 0)), F3((Sq(3), Sq(2)))))
        sage: g*f
        Module morphism:
          From: Free graded left module on 2 generators over mod 2 Steenrod algebra, milnor basis
          To:   Free graded left module on 2 generators over mod 2 Steenrod algebra, milnor basis
          Defn: b[4] |--> (Sq(0,2)+Sq(3,1)+Sq(6))*d[2]
                b[5] |--> (Sq(1,2)+Sq(7))*d[2] + (Sq(0,2)+Sq(3,1)+Sq(6))*d[3]

    TESTS:

    A non-example because the degree is not well-defined::

        sage: M = FreeGradedModule(A, (0, 0))
        sage: N = FreeGradedModule(A, (0,))
        sage: H = Hom(M, N)
        sage: g = N.generator(0)
        sage: H([Sq(1)*g, Sq(2)*g])
        Traceback (most recent call last):
        ...
        ValueError: ill-defined homomorphism: degrees do not match
    """

    def __init__(self, parent, values):
        r"""
        Initialize ``self``.

        TESTS::

            sage: from sage.modules.fp_graded.free_module import FreeGradedModule
            sage: A = SteenrodAlgebra(2)
            sage: M = FreeGradedModule(A, (0, 0))
            sage: N = FreeGradedModule(A, (0,))
            sage: H = Hom(M, N)
            sage: g = N.generator(0)
            sage: TestSuite(H).run()
            sage: TestSuite(g).run()
        """
        from .free_homspace import FreeGradedModuleHomspace
        if not isinstance(parent, FreeGradedModuleHomspace):
            raise TypeError('the parent (%s) must be a f.p. free module homset' % parent)

        self._free_morphism = self
        FPModuleMorphism.__init__(self, parent, values, check=False)

        # Compute the degree.
        if all(v.is_zero() for v in self._values):
            # The zero homomorphism does not get a degree.
            degree = None
        else:
            degrees = []
            gen_deg = parent.domain().generator_degrees()
            for i, val in enumerate(self._values):
                if val:
                    x = val.degree()
                    xx = gen_deg[i]
                    degrees.append(x - xx)

            degree = min(degrees)
            if degree != max(degrees):
                raise ValueError('ill-defined homomorphism: degrees do not match')

        self._degree = degree

    def degree(self):
        r"""
        The degree of ``self``.

        OUTPUT:

        The degree of this homomorphism. Raise an error if this is
        the trivial homomorphism.

        EXAMPLES::

            sage: from sage.modules.fp_graded.free_module import FreeGradedModule
            sage: A = SteenrodAlgebra(2)
            sage: homspace = Hom(FreeGradedModule(A, (0,1)), FreeGradedModule(A, (0,)))
            sage: N = homspace.codomain()
            sage: values = [Sq(5)*N.generator(0), Sq(3,1)*N.generator(0)]
            sage: f = homspace(values)
            sage: f.degree()
            5

        The zero homomorphism has no degree::

            sage: homspace.zero().degree()
            Traceback (most recent call last):
            ...
            ValueError: the zero morphism does not have a well-defined degree
        """
        if self._degree is None:
            # The zero morphism has no degree.
            raise ValueError("the zero morphism does not have a well-defined degree")
        return self._degree

    def __call__(self, x):
        r"""
        Evaluate the homomorphism at the given domain element ``x``.

        INPUT:

        - ``x`` -- an element of the domain of this morphism

        OUTPUT:

        The module element of the codomain which is the value of ``x``
        under this homomorphism.

        EXAMPLES::

            sage: from sage.modules.fp_graded.free_module import FreeGradedModule
            sage: A = SteenrodAlgebra(2)
            sage: M = FreeGradedModule(A, (0,1))
            sage: N = FreeGradedModule(A, (2,))
            sage: values = [Sq(5)*N.generator(0), Sq(3,1)*N.generator(0)]
            sage: f = Hom(M, N)(values)
            sage: f.__call__(M.generator(0))
            Sq(5)*g[2]
            sage: f.__call__(M.generator(1))
            Sq(3,1)*g[2]
        """
        if x.parent() != self.domain():
            raise ValueError('cannot evaluate morphism on element not in the domain')
        value = self.codomain().linear_combination(zip(self._values,
                                                       x.dense_coefficient_list()))
        return value

    def fp_module(self):
        r"""
        Create a finitely presented module from ``self``.

        OUTPUT: the finitely presented module with presentation equal to ``self``

        EXAMPLES::

            sage: A = SteenrodAlgebra(2)
            sage: F1 = A.free_graded_module([2])
            sage: F2 = A.free_graded_module([0])
            sage: v = F2([Sq(2)])
            sage: pres = Hom(F1, F2)([v])
            sage: M = pres.fp_module(); M
            Finitely presented left module on 1 generator and 1 relation over
             mod 2 Steenrod algebra, milnor basis
            sage: M.generator_degrees()
            (0,)
            sage: M.relations()
            (Sq(2)*g[0],)

            sage: from sage.modules.fp_graded.module import FPModule
            sage: A = SteenrodAlgebra(2)
            sage: F1 = A.free_graded_module((2,))
            sage: F2 = FPModule(A, (0,), [[Sq(4)]])
            sage: v = F2([Sq(2)])
            sage: pres = Hom(F1, F2)([v])
            sage: pres.fp_module()
            Traceback (most recent call last):
            ...
            ValueError: this is not a morphism between free modules
        """
        if self.codomain().has_relations():
            raise ValueError("this is not a morphism between free modules")
        try:
            FPModule = self.base_ring()._fp_graded_module_class
        except AttributeError:
            from .module import FPModule
        return FPModule(self)



# sage.doctest: needs sage.libs.flint
"""
A single element of an ambient space of modular symbols
"""
# ****************************************************************************
#       Sage: Open Source Mathematical Software
#
#       Copyright (C) 2005 William Stein <wstein@gmail.com>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  https://www.gnu.org/licenses/
# ****************************************************************************

import sage.modules.free_module_element
from sage.misc.repr import repr_lincomb
from sage.structure import formal_sum
import sage.modular.hecke.all as hecke
from sage.misc import latex


_print_mode = "manin"


def set_modsym_print_mode(mode='manin'):
    r"""
    Set the mode for printing of elements of modular symbols spaces.

    INPUT:

    - ``mode`` -- string; the possibilities are as follows:

      - ``'manin'`` -- (the default) formal sums of Manin
        symbols [P(X,Y),(u,v)]

      - ``'modular'`` -- formal sums of Modular symbols
        P(X,Y)\*alpha,beta, where alpha and beta are cusps

      - ``'vector'`` -- as vectors on the basis for the
        ambient space

    OUTPUT: none

    EXAMPLES::

        sage: M = ModularSymbols(13, 8)
        sage: x = M.0 + M.1 + M.14
        sage: set_modsym_print_mode('manin'); x
        [X^5*Y,(1,11)] + [X^5*Y,(1,12)] + [X^6,(1,11)]
        sage: set_modsym_print_mode('modular'); x
        1610510*X^6*{-1/11, 0} + 893101*X^5*Y*{-1/11, 0} + 206305*X^4*Y^2*{-1/11, 0} + 25410*X^3*Y^3*{-1/11, 0} + 1760*X^2*Y^4*{-1/11, 0} + 65*X*Y^5*{-1/11, 0} - 248832*X^6*{-1/12, 0} - 103680*X^5*Y*{-1/12, 0} - 17280*X^4*Y^2*{-1/12, 0} - 1440*X^3*Y^3*{-1/12, 0} - 60*X^2*Y^4*{-1/12, 0} - X*Y^5*{-1/12, 0} + Y^6*{-1/11, 0}
        sage: set_modsym_print_mode('vector'); x
        (1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0)
        sage: set_modsym_print_mode()
    """
    mode = str(mode).lower()
    if mode not in ['manin', 'modular', 'vector']:
        raise ValueError("mode must be one of 'manin', 'modular', or 'vector'")
    global _print_mode
    _print_mode = mode


class ModularSymbolsElement(hecke.HeckeModuleElement):
    """
    An element of a space of modular symbols.

    TESTS::

        sage: x = ModularSymbols(3, 12).cuspidal_submodule().gen(0)
        sage: x == loads(dumps(x))
        True
    """
    def __init__(self, parent, x, check=True):
        """
        INPUT:

        - ``parent`` -- a space of modular symbols

        - ``x`` -- a free module element that represents the modular
          symbol in terms of a basis for the ambient space (not in
          terms of a basis for parent!)

        EXAMPLES::

            sage: S = ModularSymbols(11, sign=1).cuspidal_submodule()
            sage: S(vector([0,1])) == S.basis()[0]
            True
            sage: S(vector([1,0]))
            Traceback (most recent call last):
            ...
            TypeError: x does not coerce to an element of this Hecke module
        """
        if check:
            from .space import ModularSymbolsSpace
            if not isinstance(parent, ModularSymbolsSpace):
                raise TypeError("parent (= %s) must be a space of modular symbols" % parent)
            if not isinstance(x, sage.modules.free_module_element.FreeModuleElement):
                raise TypeError("x must be a free module element.")
            if x.degree() != parent.degree():
                raise TypeError("x (of degree %s) must be of degree the same as the degree of the parent (of degree %s)." % (x.degree(), parent.degree()))
        hecke.HeckeModuleElement.__init__(self, parent, x)

    def _repr_(self):
        r"""
        String representation of ``self``. The output will depend on the global
        modular symbols print mode setting controlled by the function
        ``set_modsym_print_mode``.

        EXAMPLES::

            sage: M = ModularSymbols(13, 4)
            sage: set_modsym_print_mode('manin'); M.0._repr_()
            '[X^2,(0,1)]'
            sage: set_modsym_print_mode('modular'); M.0._repr_()
            'X^2*{0, Infinity}'
            sage: set_modsym_print_mode('vector'); M.0._repr_()
            '(1, 0, 0, 0, 0, 0, 0, 0)'
            sage: set_modsym_print_mode()
        """
        if _print_mode == "vector":
            return str(self.element())
        if _print_mode == "manin":
            m = self.manin_symbol_rep()
        elif _print_mode == "modular":
            m = self.modular_symbol_rep()
        return repr_lincomb([(t, c) for c, t in m])

    def _latex_(self):
        r"""
        LaTeX representation of ``self``. The output will be determined by the
        print mode setting set using ``set_modsym_print_mode``.

        EXAMPLES::

            sage: M = ModularSymbols(11, 2)
            sage: x = M.0 + M.2; x
            (1,0) + (1,9)
            sage: set_modsym_print_mode('manin'); latex(x) # indirect doctest
            (1,0) + (1,9)
            sage: set_modsym_print_mode('modular'); latex(x) # indirect doctest
            \left\{\infty, 0\right\} + \left\{\frac{-1}{9}, 0\right\}
            sage: set_modsym_print_mode('vector'); latex(x) # indirect doctest
            \left(1,\,0,\,1\right)
            sage: set_modsym_print_mode()
        """

        if _print_mode == "vector":
            return self.element()._latex_()
        if _print_mode == "manin":
            m = self.manin_symbol_rep()
        elif _print_mode == "modular":
            m = self.modular_symbol_rep()
        c = [x[0] for x in m]
        v = [x[1] for x in m]
        # TODO: use repr_lincomb with is_latex=True
        return latex.repr_lincomb(v, c)

    def _add_(self, right):
        r"""
        Sum of ``self`` and ``other``.

        EXAMPLES::

            sage: M = ModularSymbols(3, 12)
            sage: x = M.0; y = M.1; z = x + y; z # indirect doctest
            [X^8*Y^2,(1,2)] + [X^9*Y,(1,0)]
            sage: z.parent() is M
            True
        """
        return ModularSymbolsElement(self.parent(), self.element() + right.element(), check=False)

    def _rmul_(self, other):
        r"""
        Right-multiply ``self`` by ``other``.

        EXAMPLES::

            sage: M = ModularSymbols(3, 12)
            sage: x = M.0; z = x*3; z # indirect doctest
            3*[X^8*Y^2,(1,2)]
            sage: z.parent() is M
            True
            sage: z*Mod(1, 17)
            Traceback (most recent call last):
            ...
            TypeError: unsupported operand parent(s) for *: 'Modular Symbols space of dimension 8 for Gamma_0(3) of weight 12 with sign 0 over Rational Field' and 'Ring of integers modulo 17'
        """
        return ModularSymbolsElement(self.parent(), self.element() * other,
                                     check=False)

    def _lmul_(self, left):
        r"""
        Left-multiply ``self`` by ``other``.

        EXAMPLES::

            sage: M = ModularSymbols(3, 12)
            sage: x = M.0; z = 3*x; z # indirect doctest
            3*[X^8*Y^2,(1,2)]
            sage: z.parent() is M
            True
            sage: Mod(1, 17)*z
            Traceback (most recent call last):
            ...
            TypeError: unsupported operand parent(s) for *: 'Ring of integers modulo 17' and 'Modular Symbols space of dimension 8 for Gamma_0(3) of weight 12 with sign 0 over Rational Field'
        """
        return ModularSymbolsElement(self.parent(), left * self.element(),
                                     check=False)

    def _neg_(self):
        r"""
        Multiply by -1.

        EXAMPLES::

            sage: M = ModularSymbols(3, 12)
            sage: x = M.0; z = -x; z # indirect doctest
            -[X^8*Y^2,(1,2)]
            sage: z.parent() is M
            True
        """
        return ModularSymbolsElement(self.parent(), -self.element(), check=False)

    def _sub_(self, other):
        r"""
        Subtract ``other`` from ``self``.

        EXAMPLES::

            sage: M = ModularSymbols(3, 12)
            sage: x = M.0; y = M.1; z = y-x; z # indirect doctest
            -[X^8*Y^2,(1,2)] + [X^9*Y,(1,0)]
            sage: z.parent() is M
            True
        """
        return ModularSymbolsElement(self.parent(), self.element() - other.element(), check=False)

#   this clearly hasn't worked for some time -- the method embedded_vector_space doesn't exist -- DL 2009-05-18
#    def coordinate_vector(self):
#        if self.parent().is_ambient():
#            return self.element()
#        return self.parent().embedded_vector_space().coordinate_vector(self.element())

    def list(self):
        r"""
        Return a list of the coordinates of ``self`` in terms of a basis for
        the ambient space.

        EXAMPLES::

            sage: ModularSymbols(37, 2).0.list()
            [1, 0, 0, 0, 0]
        """
        return self.element().list()

    def manin_symbol_rep(self):
        """
        Return a representation of ``self`` as a formal sum of Manin symbols.

        EXAMPLES::

            sage: x = ModularSymbols(37, 4).0
            sage: x.manin_symbol_rep()
            [X^2,(0,1)]

        The result is cached::

            sage: x.manin_symbol_rep() is x.manin_symbol_rep()
            True
        """
        try:
            return self.__manin_symbols
        except AttributeError:
            A = self.parent()
            v = self.element()
            manin_symbols = A.ambient_hecke_module().manin_symbols_basis()
            F = formal_sum.FormalSums(A.base_ring())
            ms = F([(v[i], manin_symbols[i]) for i in range(v.degree())
                    if v[i] != 0], check=False, reduce=False)
            self.__manin_symbols = ms
        return self.__manin_symbols

    def modular_symbol_rep(self):
        """
        Return a representation of ``self`` as a formal sum of modular symbols.

        EXAMPLES::

            sage: x = ModularSymbols(37, 4).0
            sage: x.modular_symbol_rep()
            X^2*{0, Infinity}

        The result is cached::

            sage: x.modular_symbol_rep() is x.modular_symbol_rep()
            True
        """
        try:
            return self.__modular_symbols
        except AttributeError:
            v = self.manin_symbol_rep()
            if v == 0:
                return v
            w = [c * x.modular_symbol_rep() for c, x in v]
            self.__modular_symbols = sum(w)
            return self.__modular_symbols



r"""
Small primes of degree one

Iterator for finding several primes of absolute degree one of a number field of
*small* prime norm.

ALGORITHM:

Let `P` denote the product of some set of prime numbers. (In practice, we
use the product of the first 10000 primes, because Pari computes this many by
default.)

Let `K` be a number field and let `f(x)` be a polynomial defining `K` over the
rational field.  Let `\alpha` be a root of `f` in `K`.

We know that `[ O_K : \ZZ[\alpha] ]^2 = | \Delta(f(x)) / \Delta(O_K) |`, where
`\Delta` denotes the discriminant (see, for example, Proposition 4.4.4, p165 of
[Coh1993]_).  Therefore, after discarding primes dividing `\Delta(f(x))` (this
includes all ramified primes), any integer `n` such that `\gcd(f(n), P) > 0`
yields a prime `p | P` such that `f(x)` has a root modulo `p`.  By the
condition on discriminants, this root is a single root.  As is well known (see,
for example Theorem 4.8.13, p199 of [Coh1993]_), the ideal generated by `(p, \alpha -
n)` is prime and of degree one.

.. WARNING::

    It is possible that there are no primes of `K` of absolute degree one of
    small prime norm, and it is possible that this algorithm will not find
    any primes of small norm.

.. TODO::

    There are situations when this will fail.  There are questions of finding
    primes of relative degree one.  There are questions of finding primes of exact
    degree larger than one.  In short, if you can contribute, please do!

EXAMPLES::

    sage: x = ZZ['x'].gen()
    sage: F.<a> = NumberField(x^2 - 2)
    sage: Ps = F.primes_of_degree_one_list(3)
    sage: Ps # random
    [Fractional ideal (2*a + 1), Fractional ideal (-3*a + 1), Fractional ideal (-a + 5)]
    sage: [ P.norm() for P in Ps ] # random
    [7, 17, 23]
    sage: all(ZZ(P.norm()).is_prime() for P in Ps)
    True
    sage: all(P.residue_class_degree() == 1 for P in Ps)
    True

The next two examples are for relative number fields.::

    sage: L.<b> = F.extension(x^3 - a)
    sage: Ps = L.primes_of_degree_one_list(3)
    sage: Ps # random
    [Fractional ideal (17, b - 5), Fractional ideal (23, b - 4), Fractional ideal (31, b - 2)]
    sage: [ P.absolute_norm() for P in Ps ] # random
    [17, 23, 31]
    sage: all(ZZ(P.absolute_norm()).is_prime() for P in Ps)
    True
    sage: all(P.residue_class_degree() == 1 for P in Ps)
    True
    sage: M.<c> = NumberField(x^2 - x*b^2 + b)
    sage: Ps = M.primes_of_degree_one_list(3)
    sage: Ps # random
    [Fractional ideal (17, c - 2), Fractional ideal (c - 1), Fractional ideal (41, c + 15)]
    sage: [ P.absolute_norm() for P in Ps ] # random
    [17, 31, 41]
    sage: all(ZZ(P.absolute_norm()).is_prime() for P in Ps)
    True
    sage: all(P.residue_class_degree() == 1 for P in Ps)
    True

AUTHORS:

- Nick Alexander (2008): initial version
- David Loeffler (2009): fixed a bug with relative fields
- Maarten Derickx (2017): fixed a bug with number fields not generated by an integral element
"""

#*****************************************************************************
#       Copyright (C) 2008 William Stein
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  http://www.gnu.org/licenses/
#*****************************************************************************

from sage.rings.integer_ring import ZZ


class Small_primes_of_degree_one_iter:
    r"""
    Iterator that finds primes of a number field of absolute degree
    one and bounded small prime norm.

    INPUT:

    - ``field`` -- a :class:`NumberField`

    - ``num_integer_primes`` -- integer (default: 10000); we try to find
      primes of absolute norm no greater than the ``num_integer_primes``-th
      prime number. For example, if ``num_integer_primes`` is 2, the largest
      norm found will be 3, since the second prime is 3.

    - ``max_iterations`` -- integer (default: 100); we test ``max_iterations``
      integers to find small primes before raising :class:`StopIteration`

    AUTHOR:

    - Nick Alexander
    """
    def __init__(self, field, num_integer_primes=10000, max_iterations=100):
        r"""
        Construct a new iterator of small degree one primes.

        EXAMPLES::

            sage: x = QQ['x'].gen()
            sage: K.<a> = NumberField(x^2 - 3)
            sage: K.primes_of_degree_one_list(3) # random
            [Fractional ideal (2*a + 1), Fractional ideal (-a + 4), Fractional ideal (3*a + 2)]
        """
        self._field = field
        self._poly = self._field.absolute_field('b').defining_polynomial()
        self._poly = ZZ['x'](self._poly.denominator() * self._poly()) # make integer polynomial
        self._lc = self._poly.leading_coefficient()

        # this uses that [ O_K : Z[a] ]^2 = | disc(f(x)) / disc(O_K) |
        from sage.libs.pari import pari
        self._prod_of_small_primes = ZZ(pari('TEMPn = %s; TEMPps = primes(TEMPn); prod(X = 1, TEMPn, TEMPps[X])' % num_integer_primes))
        self._prod_of_small_primes //= self._prod_of_small_primes.gcd(self._poly.discriminant() * self._lc)

        self._integer_iter = iter(ZZ)
        self._queue = []
        self._max_iterations = max_iterations

    def __iter__(self):
        r"""
        Return ``self`` as an iterator.

        EXAMPLES::

            sage: x = QQ['x'].gen()
            sage: K.<a> = NumberField(x^2 - 3)
            sage: it = K.primes_of_degree_one_iter()
            sage: iter(it) == it # indirect doctest
            True
        """
        return self

    def _lengthen_queue(self):
        r"""
        Try to find more primes of absolute degree one of small prime
        norm.

        Checks \code{self._max_iterations} integers before failing.

        WARNING:

            Internal function.  Not for external use!

        EXAMPLES::

            sage: x = QQ['x'].gen()
            sage: K.<a> = NumberField(x^2 - 3)
            sage: Ps = K.primes_of_degree_one_list(20, max_iterations=3) # indirect doctest
            sage: len(Ps) == 20
            True
        """
        count = 0
        while count < self._max_iterations:
            n = next(self._integer_iter)
            g = self._prod_of_small_primes.gcd(self._poly(n))
            self._prod_of_small_primes //= g
            self._queue = self._queue + [ (p, n) for p in g.prime_divisors() ]
            count += 1
        self._queue.sort() # sorts in ascending order

    def __next__(self):
        r"""
        Return a prime of absolute degree one of small prime norm.

        Raises ``StopIteration`` if such a prime cannot be easily found.

        EXAMPLES::

            sage: x = QQ['x'].gen()
            sage: K.<a> = NumberField(x^2 - 3)
            sage: it = K.primes_of_degree_one_iter()
            sage: [ next(it) for i in range(3) ] # random
            [Fractional ideal (2*a + 1), Fractional ideal (-a + 4), Fractional ideal (3*a + 2)]

        TESTS:

        We test that :issue:`6396` is fixed. Note that the doctest is
        flagged as random since the string representation of ideals is
        somewhat unpredictable::

            sage: N.<a,b> = NumberField([x^2 + 1, x^2 - 5])
            sage: ids = N.primes_of_degree_one_list(10); ids  # random
            [Fractional ideal ((-1/2*b + 1/2)*a + 2),
             Fractional ideal (-b*a + 1/2*b + 1/2),
             Fractional ideal ((1/2*b + 3/2)*a - b),
             Fractional ideal ((-1/2*b - 3/2)*a + b - 1),
             Fractional ideal (-b*a - b + 1),
             Fractional ideal (3*a + 1/2*b - 1/2),
             Fractional ideal ((-3/2*b + 1/2)*a + 1/2*b - 1/2),
             Fractional ideal ((-1/2*b - 5/2)*a - b + 1),
             Fractional ideal (2*a - 3/2*b - 1/2),
             Fractional ideal (3*a + 1/2*b + 5/2)]
             sage: [x.absolute_norm() for x in ids]
             [29, 41, 61, 89, 101, 109, 149, 181, 229, 241]
             sage: ids[9] == N.ideal(3*a + 1/2*b + 5/2)
             True

        We test that :issue:`23468` is fixed::

            sage: R.<z> = QQ[]
            sage: K.<y> = QQ.extension(25*z^2 + 26*z + 5)
            sage: for p in K.primes_of_degree_one_list(10):
            ....:     assert p.is_prime()
        """
        if not self._queue:
            self._lengthen_queue()
        if not self._queue:
            raise StopIteration

        p, n = self._queue.pop(0)
        x = self._field.absolute_generator()
        return self._field.ideal([p, (x - n) * self._lc])

    next = __next__



"""
`q`-expansions of theta series

AUTHOR:

- William Stein
"""
from sage.rings.integer import Integer
from sage.rings.integer_ring import ZZ
from sage.rings.power_series_ring import PowerSeriesRing

from math import sqrt


def theta2_qexp(prec=10, var='q', K=ZZ, sparse=False):
    r"""
    Return the `q`-expansion of the series `\theta_2 = \sum_{n \text{ odd}} q^{n^2}`.

    INPUT:

    - ``prec`` -- integer; the absolute precision of the output
    - ``var`` -- (default: ``'q'``) variable name
    - ``K`` -- (default: ZZ) base ring of answer

    OUTPUT: a power series over K

    EXAMPLES::

        sage: theta2_qexp(18)
        q + q^9 + O(q^18)
        sage: theta2_qexp(49)
        q + q^9 + q^25 + O(q^49)
        sage: theta2_qexp(100, 'q', QQ)
        q + q^9 + q^25 + q^49 + q^81 + O(q^100)
        sage: f = theta2_qexp(100, 't', GF(3)); f
        t + t^9 + t^25 + t^49 + t^81 + O(t^100)
        sage: parent(f)
        Power Series Ring in t over Finite Field of size 3
        sage: theta2_qexp(200)
        q + q^9 + q^25 + q^49 + q^81 + q^121 + q^169 + O(q^200)
        sage: f = theta2_qexp(20,sparse=True); f
        q + q^9 + O(q^20)
        sage: parent(f)
        Sparse Power Series Ring in q over Integer Ring
    """
    prec = Integer(prec)
    if prec <= 0:
        raise ValueError("prec must be positive")
    if sparse:
        v = {}
    else:
        v = [Integer(0)] * prec
    one = Integer(1)
    n = int(sqrt(prec))
    if n*n < prec:
        n += 1
    for m in range(1, n, 2):
        v[m*m] = one
    R = PowerSeriesRing(K, sparse=sparse, names=var)
    return R(v, prec=prec)


def theta_qexp(prec=10, var='q', K=ZZ, sparse=False):
    r"""
    Return the `q`-expansion of the standard `\theta` series
    `\theta = 1 + 2\sum_{n=1}^{\infty} q^{n^2}`.

    INPUT:

    - ``prec`` -- integer; the absolute precision of the output
    - ``var`` -- (default: ``'q'``) variable name
    - ``K`` -- (default: ZZ) base ring of answer

    OUTPUT: a power series over K

    EXAMPLES::

        sage: theta_qexp(25)
        1 + 2*q + 2*q^4 + 2*q^9 + 2*q^16 + O(q^25)
        sage: theta_qexp(10)
        1 + 2*q + 2*q^4 + 2*q^9 + O(q^10)
        sage: theta_qexp(100)
        1 + 2*q + 2*q^4 + 2*q^9 + 2*q^16 + 2*q^25 + 2*q^36 + 2*q^49 + 2*q^64 + 2*q^81 + O(q^100)
        sage: theta_qexp(100, 't')
        1 + 2*t + 2*t^4 + 2*t^9 + 2*t^16 + 2*t^25 + 2*t^36 + 2*t^49 + 2*t^64 + 2*t^81 + O(t^100)
        sage: theta_qexp(100, 't', GF(2))
        1 + O(t^100)
        sage: f = theta_qexp(20,sparse=True); f
        1 + 2*q + 2*q^4 + 2*q^9 + 2*q^16 + O(q^20)
        sage: parent(f)
        Sparse Power Series Ring in q over Integer Ring
    """
    prec = Integer(prec)
    if prec <= 0:
        raise ValueError("prec must be positive")
    if sparse:
        v = {}
    else:
        v = [Integer(0)] * prec
    v[0] = Integer(1)
    two = Integer(2)
    n = int(sqrt(prec))
    if n*n != prec:
        n += 1
    for m in range(1, n):
        v[m*m] = two

    R = PowerSeriesRing(K, sparse=sparse, names=var)
    return R(v, prec=prec)



"""
Necklaces

The algorithm used in this file comes from

- Sawada, Joe. *A fast algorithm to generate necklaces with fixed content*,
  Theoretical Computer Science archive Volume 301, Issue 1-3 (May 2003)
  :doi:`10.1016/S0304-3975(03)00049-5`
"""
# ****************************************************************************
#       Copyright (C) 2007 Mike Hansen <mhansen@gmail.com>,
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from sage.arith.misc import divisors, euler_phi, factorial, gcd
from sage.categories.finite_enumerated_sets import FiniteEnumeratedSets
from sage.combinat.composition import Composition
from sage.combinat.misc import DoublyLinkedList
from sage.misc.misc_c import prod
from sage.rings.integer import Integer
from sage.rings.integer_ring import ZZ
from sage.structure.parent import Parent
from sage.structure.unique_representation import UniqueRepresentation


def Necklaces(content):
    r"""
    Return the set of necklaces with evaluation ``content``.

    A necklace is a list of integers that such that the list is
    the smallest lexicographic representative of all the cyclic shifts
    of the list.

    .. SEEALSO::

        :class:`LyndonWords`

    INPUT:

    - ``content`` -- list or tuple of nonnegative integers

    EXAMPLES::

        sage: Necklaces([2,1,1])
        Necklaces with evaluation [2, 1, 1]
        sage: Necklaces([2,1,1]).cardinality()
        3
        sage: Necklaces([2,1,1]).first()
        [1, 1, 2, 3]
        sage: Necklaces([2,1,1]).last()
        [1, 2, 1, 3]
        sage: Necklaces([2,1,1]).list()
        [[1, 1, 2, 3], [1, 1, 3, 2], [1, 2, 1, 3]]
        sage: Necklaces([0,2,1,1]).list()
        [[2, 2, 3, 4], [2, 2, 4, 3], [2, 3, 2, 4]]
        sage: Necklaces([2,0,1,1]).list()
        [[1, 1, 3, 4], [1, 1, 4, 3], [1, 3, 1, 4]]
    """
    return Necklaces_evaluation(content)


class Necklaces_evaluation(UniqueRepresentation, Parent):
    """
    Necklaces with a fixed evaluation (content).

    INPUT:

    - ``content`` -- list or tuple of nonnegative integers
    """
    @staticmethod
    def __classcall_private__(cls, content):
        """
        Return the correct parent object, with standardized parameters.

        EXAMPLES::

            sage: Necklaces([2,1,1]) is Necklaces(Composition([2,1,1]))
            True
        """
        if not isinstance(content, Composition):
            content = Composition(content)
        return super().__classcall__(cls, content)

    def __init__(self, content):
        r"""
        Initialize ``self``.

        TESTS::

            sage: N = Necklaces([2,2,2])
            sage: N == loads(dumps(N))
            True
            sage: T = Necklaces([2,1])
            sage: TestSuite(T).run()
        """
        self._content = content
        Parent.__init__(self, category=FiniteEnumeratedSets())

    def content(self):
        """
        Return the content (or evaluation) of the necklaces.

        EXAMPLES::

            sage: N = Necklaces([2,2,2])
            sage: N.content()
            [2, 2, 2]
        """
        return self._content

    def __repr__(self) -> str:
        r"""
        TESTS::

            sage: repr(Necklaces([2,1,1]))
            'Necklaces with evaluation [2, 1, 1]'
        """
        return "Necklaces with evaluation %s" % self._content

    def __contains__(self, x) -> bool:
        r"""
        Return ``True`` if ``x`` is the smallest word of all its cyclic shifts
        and the content vector of ``x`` is equal to ``content``.

        INPUT:

        - ``x`` -- list of integers

        EXAMPLES::

            sage: [2,1,2,1] in Necklaces([2,2])
            False
            sage: [1,2,1,2] in Necklaces([2,2])
            True
            sage: [1,1,2,2] in Necklaces([2,2])
            True
            sage: [1,2,2,2] in Necklaces([2,2])
            False
            sage: all(n in Necklaces([2,1,3,1]) for n in Necklaces([2,1,3,1]))
            True
            sage: all(n in Necklaces([0,1,2,3]) for n in Necklaces([0,1,2,3]))
            True
        """
        xl = list(x)
        e = [0] * len(self._content)
        if len(xl) != sum(self._content):
            return False

        # Check to make sure xl is a list of integers
        for i in xl:
            if not isinstance(i, (int, Integer)):
                return False
            if i <= 0:
                return False
            if i > len(self._content):
                return False
            e[i - 1] += 1

        # Check to make sure the evaluation is the same
        if e != self._content:
            return False

        # Check to make sure that x is lexicographically less
        # than all of its cyclic shifts
        cyclic_shift = xl[:]
        for i in range(len(xl) - 1):
            cyclic_shift = cyclic_shift[1:] + cyclic_shift[:1]
            if cyclic_shift < xl:
                return False

        return True

    def cardinality(self) -> Integer:
        r"""
        Return the number of integer necklaces with the evaluation ``content``.

        The formula for the number of necklaces of content `\alpha`
        a composition of `n` is:

        .. MATH::

            \sum_{d|gcd(\alpha)} \phi(d)
            \binom{n/d}{\alpha_1/d, \ldots, \alpha_\ell/d},

        where `\phi(d)` is the Euler `\phi` function.

        EXAMPLES::

            sage: Necklaces([]).cardinality()
            0
            sage: Necklaces([2,2]).cardinality()
            2
            sage: Necklaces([2,3,2]).cardinality()
            30
            sage: Necklaces([0,3,2]).cardinality()
            2

        Check to make sure that the count matches up with the number of
        necklace words generated.

        ::

            sage: comps = [[],[2,2],[3,2,7],[4,2],[0,4,2],[2,0,4]] + Compositions(4).list()
            sage: ns = [Necklaces(comp) for comp in comps]
            sage: all(n.cardinality() == len(n.list()) for n in ns)                     # needs sage.libs.pari
            True
        """
        evaluation = self._content
        le = list(evaluation)
        if not le:
            return ZZ.zero()

        n = sum(le)

        return ZZ.sum(euler_phi(j) * factorial(n // j) //
                      prod(factorial(ni // j) for ni in evaluation)
                      for j in divisors(gcd(le))) // n

    def __iter__(self):
        r"""
        An iterator for the integer necklaces with evaluation ``content``.

        EXAMPLES::

            sage: Necklaces([]).list()    #indirect test
            []
            sage: Necklaces([1]).list()   #indirect test
            [[1]]
            sage: Necklaces([2]).list()   #indirect test
            [[1, 1]]
            sage: Necklaces([3]).list()   #indirect test
            [[1, 1, 1]]
            sage: Necklaces([3,3]).list() #indirect test
            [[1, 1, 1, 2, 2, 2],
             [1, 1, 2, 1, 2, 2],
             [1, 1, 2, 2, 1, 2],
             [1, 2, 1, 2, 1, 2]]
            sage: Necklaces([2,1,3]).list() #indirect test
            [[1, 1, 2, 3, 3, 3],
             [1, 1, 3, 2, 3, 3],
             [1, 1, 3, 3, 2, 3],
             [1, 1, 3, 3, 3, 2],
             [1, 2, 1, 3, 3, 3],
             [1, 2, 3, 1, 3, 3],
             [1, 2, 3, 3, 1, 3],
             [1, 3, 1, 3, 2, 3],
             [1, 3, 1, 3, 3, 2],
             [1, 3, 2, 1, 3, 3]]
        """
        if not self._content:
            return
        k = 0
        while not self._content[k]:  # == 0
            k += 1
        for z in _sfc(self._content[k:]):
            yield [x + 1 + k for x in z]


################################
# Fast Fixed Content Algorithm #
################################
def _ffc(content, equality=False):
    """
    EXAMPLES::

        sage: from sage.combinat.necklace import _ffc
        sage: list(_ffc([3,3])) #necklaces
        [[0, 1, 0, 1, 0, 1],
         [0, 0, 1, 1, 0, 1],
         [0, 0, 1, 0, 1, 1],
         [0, 0, 0, 1, 1, 1]]
        sage: list(_ffc([3,3], equality=True)) #Lyndon words
        [[0, 0, 1, 1, 0, 1], [0, 0, 1, 0, 1, 1], [0, 0, 0, 1, 1, 1]]
    """
    e = list(content)
    a = [len(e) - 1] * sum(e)
    r = [0] * sum(e)
    a[0] = 0
    e[0] -= 1
    k = len(e)

    rng_k = list(range(k))
    rng_k.reverse()
    dll = DoublyLinkedList(rng_k)
    if not e[0]:  # == 0
        dll.hide(0)

    yield from _fast_fixed_content(a, e, 2, 1, k, r, 2, dll, equality=equality)


def _fast_fixed_content(a, content, t, p, k, r, s, dll, equality=False):
    """
    EXAMPLES::

        sage: from sage.combinat.necklace import _fast_fixed_content
        sage: from sage.combinat.misc import DoublyLinkedList
        sage: e = [3,3]
        sage: a = [len(e)-1]*sum(e)
        sage: r = [0]*sum(e)
        sage: a[0] = 0
        sage: e[0] -= 1
        sage: k = len(e)
        sage: dll = DoublyLinkedList(list(reversed(range(k))))
        sage: if e[0] == 0: dll.hide(0)
        sage: list(_fast_fixed_content(a,e,2,1,k,r,2,dll))
        [[0, 1, 0, 1, 0, 1],
         [0, 0, 1, 1, 0, 1],
         [0, 0, 1, 0, 1, 1],
         [0, 0, 0, 1, 1, 1]]
        sage: list(_fast_fixed_content(a,e,2,1,k,r,2,dll,True))
        [[0, 0, 1, 1, 0, 1], [0, 0, 1, 0, 1, 1], [0, 0, 0, 1, 1, 1]]
    """
    n = len(a)
    if content[k - 1] == n - t + 1:
        if content[k - 1] == r[t - p - 1]:
            if equality:
                if n == p:
                    yield a
            else:
                if not n % p:  # == 0
                    yield a
        elif content[k - 1] > r[t - p - 1]:
            yield a
    elif content[0] != n - t + 1:
        j = dll.head()
        sp = s
        while j != 'end' and j >= a[t - p - 1]:
            r[s - 1] = t - s
            a[t - 1] = j
            content[j] -= 1

            if not content[j]:  # == 0
                dll.hide(j)

            if j != k - 1:
                sp = t + 1

            if j == a[t - p - 1]:
                yield from _fast_fixed_content(a[:], content, t + 1, p,
                                               k, r, sp, dll,
                                               equality=equality)
            else:
                yield from _fast_fixed_content(a[:], content, t + 1, t,
                                               k, r, sp, dll,
                                               equality=equality)

            if not content[j]:  # == 0
                dll.unhide(j)

            content[j] += 1
            j = dll.next(j)
        a[t - 1] = k - 1


# ###############################
# List Fixed Content Algorithm  #
# ###############################
def _lfc(content, equality=False):
    """
    EXAMPLES::

        sage: from sage.combinat.necklace import _lfc
        sage: list(_lfc([3,3])) #necklaces
        [[0, 1, 0, 1, 0, 1],
         [0, 0, 1, 1, 0, 1],
         [0, 0, 1, 0, 1, 1],
         [0, 0, 0, 1, 1, 1]]
        sage: list(_lfc([3,3], equality=True)) #Lyndon words
        [[0, 0, 1, 1, 0, 1], [0, 0, 1, 0, 1, 1], [0, 0, 0, 1, 1, 1]]
    """
    content = list(content)
    a = [0] * sum(content)
    content[0] -= 1
    k = len(content)

    rng_k = list(range(k))
    rng_k.reverse()
    dll = DoublyLinkedList(rng_k)

    if not content[0]:  # == 0
        dll.hide(0)

    yield from _list_fixed_content(a, content, 2, 1, k, dll, equality=equality)


def _list_fixed_content(a, content, t, p, k, dll, equality=False):
    """
    EXAMPLES::

        sage: from sage.combinat.necklace import _list_fixed_content
        sage: from sage.combinat.misc import DoublyLinkedList
        sage: e = [3,3]
        sage: a = [0]*sum(e)
        sage: e[0] -= 1
        sage: k = len(e)
        sage: dll = DoublyLinkedList(list(reversed(range(k))))
        sage: if e[0] == 0: dll.hide(0)
        sage: list(_list_fixed_content(a,e,2,1,k,dll))
        [[0, 1, 0, 1, 0, 1],
         [0, 0, 1, 1, 0, 1],
         [0, 0, 1, 0, 1, 1],
         [0, 0, 0, 1, 1, 1]]
        sage: list(_list_fixed_content(a,e,2,1,k,dll,True))
        [[0, 0, 1, 1, 0, 1], [0, 0, 1, 0, 1, 1], [0, 0, 0, 1, 1, 1]]
    """
    n = len(a)
    if t > n:
        if equality:
            if n == p:
                yield a
        else:
            if not n % p:  # == 0
                yield a
    else:
        j = dll.head()
        while j != 'end' and j >= a[t - p - 1]:
            a[t - 1] = j
            content[j] -= 1

            if not content[j]:  # == 0
                dll.hide(j)

            if j == a[t - p - 1]:
                yield from _list_fixed_content(a[:], content[:], t + 1, p,
                                               k, dll, equality=equality)
            else:
                yield from _list_fixed_content(a[:], content[:], t + 1, t,
                                               k, dll, equality=equality)

            if not content[j]:  # == 0
                dll.unhide(j)

            content[j] += 1
            j = dll.next(j)


##################################
# Simple Fixed Content Algorithm #
##################################
def _sfc(content, equality=False):
    """
    This wrapper function calls :meth:`sage.combinat.necklace._simple_fixed_content`.
    If ``equality`` is ``True`` the function returns Lyndon words with content
    vector equal to ``content``, otherwise it returns necklaces.

    INPUT:

    - ``content`` -- list of nonnegative integers with no leading 0s
    - ``equality`` -- boolean (default: ``True``)

    .. WARNING::

        You will get incorrect results if there are leading 0s in ``content``.
        See :issue:`12997` and :issue:`17436`.

    EXAMPLES::

        sage: from sage.combinat.necklace import _sfc
        sage: list(_sfc([3,3])) #necklaces
        [[0, 0, 0, 1, 1, 1],
         [0, 0, 1, 0, 1, 1],
         [0, 0, 1, 1, 0, 1],
         [0, 1, 0, 1, 0, 1]]
        sage: list(_sfc([3,3], equality=True)) #Lyndon words
        [[0, 0, 0, 1, 1, 1], [0, 0, 1, 0, 1, 1], [0, 0, 1, 1, 0, 1]]
    """
    content = list(content)
    a = [0] * sum(content)
    content[0] -= 1
    k = len(content)
    return _simple_fixed_content(a, content, 2, 1, k, equality=equality)


def _simple_fixed_content(a, content, t, p, k, equality=False):
    """
    EXAMPLES::

        sage: from sage.combinat.necklace import _simple_fixed_content
        sage: content = [3,3]
        sage: a = [0]*sum(content)
        sage: content[0] -= 1
        sage: k = len(content); k
        2
        sage: list(_simple_fixed_content(a, content, 2, 1, k))
        [[0, 0, 0, 1, 1, 1],
         [0, 0, 1, 0, 1, 1],
         [0, 0, 1, 1, 0, 1],
         [0, 1, 0, 1, 0, 1]]
        sage: list(_simple_fixed_content(a, content, 2, 1, k, True))
        [[0, 0, 0, 1, 1, 1], [0, 0, 1, 0, 1, 1], [0, 0, 1, 1, 0, 1]]
    """
    n = len(a)
    if t > n:
        if equality:
            if n == p:
                yield a
        else:
            if not n % p:  # == 0
                yield a
    else:
        r = list(range(a[t - p - 1], k))
        for j in r:
            if content[j] > 0:
                a[t - 1] = j
                content[j] -= 1
                if j == a[t - p - 1]:
                    yield from _simple_fixed_content(a[:], content, t + 1, p,
                                                     k, equality=equality)
                else:
                    yield from _simple_fixed_content(a[:], content, t + 1, t,
                                                     k, equality=equality)
                content[j] += 1


def _lyn(w):
    """
    Return the length of the longest prefix of ``w`` that is a Lyndon word.

    EXAMPLES::

        sage: import sage.combinat.necklace as necklace
        sage: necklace._lyn([0,1,1,0,0,1,2])
        3
        sage: necklace._lyn([0,0,0,1])
        4
        sage: necklace._lyn([2,1,0,0,2,2,1])
        1
    """
    p = 1
    k = max(w) + 1
    for i in range(1, len(w)):
        b = w[i]
        a = w[:i]
        if b < a[i - p] or b > k - 1:
            return p
        if b == a[i - p]:
            pass
        else:
            p = i + 1
    return p
